/*
 * tclGeoPlace - maintain a database of named places 
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 ********************************************
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <tcl.h>
#include "geography.h"
#include "tclGeoPlace.h"

#define LENGTH 255
#define GEOPLACE "geoplace"

/* Structure for a named place */
typedef struct
{
  GeoPoint latLon;		/* Latitude and longitude of a place */
  unsigned userCnt;		/* Number of widgets using the place */
} GeoPlace;

static int loaded;
static char result[200];	/* Hold a result */

static Tcl_HashTable placeCmds;	/* Map place commands to their functions */
static Tcl_HashTable places;	/* Table to map place names to locations */
static Tcl_HashTable updateCmds;/* Table of cmds to run when places changes */

static Tcl_CmdProc geoPlaceCmd;	/* Parse geoPlace command */
static Tcl_CmdProc newPlace;	/* Create a new place */
static Tcl_CmdProc setPlace;	/* Set or report location for a place */
static Tcl_CmdProc updateCmd;	/* Identify command to run if a place changes */
static Tcl_CmdProc rmPlace;	/* Remove a place */
static Tcl_CmdProc listPlaces;	/* Return a list of place names */
static Tcl_CmdProc distance;	/* Compute distance between two places */
static Tcl_CmdProc nearest;	/* Find place nearest a point */
static Tcl_CmdProc step;	/* Step a given distance and range from point */
static Tcl_CmdProc inrange;	/* Find places within range of given point */
static Tcl_CmdProc lonbtwn;	/* Determine if a lon is between two others */

static void 
runUpdateCmds(Tcl_Interp *interp); /* Run commands if a geoplace changes */

int
Tclgeoplace_Init(Tcl_Interp *interp)
{
  /*
   * GeoPlacesInit - create the table to map place names to locations and
   * create the commands to access it.
   * Input: interp - tcl interpreter
   * Return TCL_OK or TCL_ERROR
   */

  /* Struct to map command name to command function */
  struct NamedCommand
  {
    char		*name;
    Tcl_CmdProc   *function;
  };

  /* GeoPlace commands and their functions */
  struct NamedCommand geoPlaceCmdList[] =
    {
      {"new",		newPlace },
      {"set",		setPlace },
      {"updatecmd",	updateCmd },
      {"rm",		rmPlace },
      {"remove",	rmPlace },
      {"list",		listPlaces },
      {"distance", 	distance },
      {"nearest", 	nearest },
      {"step", 		step},
      {"inrange", 	inrange},
      {"lonbtwn", 	lonbtwn},
      {NULL,		NULL }
    };
  struct NamedCommand *command;	/* Loop pointer */
  int np;

  if (loaded)
    return TCL_OK;
  loaded = 1;

  /* Create place command and subcommands */
  Tcl_CreateCommand(interp, GEOPLACE, geoPlaceCmd, NULL,
		    (Tcl_CmdDeleteProc *)NULL);
  Tcl_InitHashTable(&placeCmds, TCL_STRING_KEYS);
  for (command = geoPlaceCmdList; command->name; command++)
    Tcl_SetHashValue(Tcl_CreateHashEntry(&placeCmds, command->name, &np),
		     (ClientData)command->function);

  /* Initiailize the table of named places */
  Tcl_InitHashTable(&places, TCL_STRING_KEYS);

  /* Initialize table of commands to run if a place changes */
  Tcl_InitHashTable(&updateCmds, TCL_STRING_KEYS);

  Tcl_PkgProvide(interp, "tclgeoplace", "1.2");

  return TCL_OK;
}

int
geoPlaceCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  /* geoPlaceCmd - called for GEOPLACE on command line.  Find and execute 
   * function corresponding to first argument.
   * Usage:
   *   geoplace command ?option? ?option? ...
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   Result of executed command, or TCL_ERROR.
   * Side effects:
   *   Appends error message to interp->result on error
   */

  int result;
  Tcl_HashEntry *entry;
  Tcl_CmdProc *command;
  Tcl_HashSearch search;

  if ( (argc == 1) || !(entry = Tcl_FindHashEntry(&placeCmds, argv[1])) ) {

    /* No command given or given command not found => give a list of commands */
    Tcl_AppendResult(interp, argv[0], " commands are: ", (char *)NULL);
    for (entry = Tcl_FirstHashEntry(&placeCmds, &search);
         entry != NULL;
         entry = Tcl_NextHashEntry(&search))
      Tcl_AppendResult(interp, Tcl_GetHashKey(&placeCmds, entry), " ",
			 (char*)NULL);
    result = TCL_ERROR;

  } else {

    /* Execute the geoplace command */
    command = (Tcl_CmdProc *)Tcl_GetHashValue(entry);
    result = (*command)(NULL, interp, argc, argv);

  } 

  return result;
}

int
GeoPlaceGet(Tcl_Interp *interp, char *placeName, GeoPoint *loc,	int useFlag)
{
  /* Find table entry corresponding to key placeName, put its location
   * into loc.  If useFlag is nonzero, increment it's usercount.
   * Normally returns TCL_OK.
   * If place does not exist, puts an error message into interp->result
   * and returns TCL_ERROR */

  Tcl_HashEntry *entry;
  GeoPlace *place;

  if ((entry = Tcl_FindHashEntry(&places, placeName)) == NULL)
    {
      Tcl_AppendResult(interp, "No place named ", placeName, (char *)NULL);
      return TCL_ERROR;
    }
  place = Tcl_GetHashValue(entry);
  *loc = place->latLon;
  if (useFlag == KEEP_PLACE)
    place->userCnt++;
  return TCL_OK;
}

void
GeoPlaceFree(char *placeName)
{
  /* GeoPlaceFree - indicate no further use for a place identified by key
   * placeName */
  
  Tcl_HashEntry *entry;
  GeoPlace *place;

  if ((entry = Tcl_FindHashEntry(&places, placeName)) == NULL)
    return;
  place = Tcl_GetHashValue(entry);
  place->userCnt--;
}

int
newPlace(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* newPlace - create a new place of interest
   * Usage:
   *   geoplace new placeName {lat lon}
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   Result of executed command, or TCL_ERROR.
   * Side effects:
   *   Appends error message to interp->result on error.
   */

  char
    *placeName,
    *location;
  Tcl_HashEntry *entry;		/* Entry for place */
  int newPtr;			/* Indicate if place exists*/
  double lat, lon;		/* Lat and lon for place */
  GeoPlace *place;		/* Place of interest */

  if (argc != 4)
    {
      Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		       " placeName {lat lon}", (char *)NULL);
      return TCL_ERROR;
    } 

  /* Get latitude and longitude of the new place */
  location = argv[3];
  if (sscanf(location, "%lf %lf", &lat, &lon) != 2) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " expected {lat lon}, "
      "got ", location, (char *)NULL);
    return TCL_ERROR;
  }

  /* Add the new place to the database */
  placeName = argv[2];
  entry = Tcl_CreateHashEntry(&places, placeName, &newPtr);
  if (newPtr == 0) {
    /* Allocate and initialize new place */
    Tcl_AppendResult(interp, placeName, " already exists", (char *)NULL);
    return TCL_ERROR;
  }
  if ((place = (GeoPlace *)malloc(sizeof(GeoPlace))) == NULL) {
    Tcl_DeleteHashEntry(entry);
    Tcl_AppendResult(interp, "Allocation failed for ", placeName, (char *)NULL);
    return TCL_ERROR;
  }
  place->userCnt = 0;
  place->latLon.ord = lat;
  place->latLon.abs = lon;
  Tcl_SetHashValue(entry, place);
  Tcl_SetResult(interp, placeName, TCL_VOLATILE);
  runUpdateCmds(interp);
  return TCL_OK;
}

int
setPlace(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* setPlace - set location for a place, creating it if necessary, or return
   * name, latitude and longitude of a place
   * Usage:
   *   geoplace set placeName ?{lat lon}?
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Appends place coordinates to interp->result if no new coordinates given
   *   Appends error message to interp->result on failure
   */

  char
    *placeName,
    *location;
  Tcl_HashEntry *entry;		/* Entry for place */
  int newPtr;			/* Indicate if place exists*/
  double lat, lon;		/* Lat and lon for place */
  GeoPlace *place;		/* Place of interest */

  if (argc == 3) {
    /* Return place location */
    placeName = argv[2];
    if ( !(entry = Tcl_FindHashEntry(&places, placeName)) ) {
      Tcl_AppendResult(interp, "No place named ", placeName, (char *)NULL);
      return TCL_ERROR;
    } else {
      place = Tcl_GetHashValue(entry);
      sprintf(result, "%8.2f %8.2f", place->latLon.ord, place->latLon.abs);
      Tcl_SetResult(interp, result, TCL_VOLATILE);
    }
  } else if (argc == 4) {
    /* Set location */
    placeName = argv[2];
    location = argv[3];

    /* Read in desired {lat lon} */
    if (sscanf(location, "%lf %lf", &lat, &lon) != 2) {
      Tcl_AppendResult(interp, "Expected {lat lon}, got ", location,
        (char *)NULL);
      return TCL_ERROR;
    }

    /* Find or create place in the GeoPlace database */
    entry = Tcl_CreateHashEntry(&places, placeName, &newPtr);
    if (newPtr == 1) {
      /* Allocate and initialize new place */
      if ((place = (GeoPlace *)malloc(sizeof(GeoPlace))) == NULL) {
        Tcl_DeleteHashEntry(entry);
        Tcl_AppendResult(interp, "Allocation failed for ", placeName,
          (char *)NULL);
        return TCL_ERROR;
      }
      place->userCnt = 0;
    } else {
      /* Get struct for already existing place */
      place = Tcl_GetHashValue(entry);
    }

    /* Set place location */
    place->latLon.ord = lat;
    place->latLon.abs = lon;
    Tcl_SetHashValue(entry, place);
    Tcl_SetResult(interp, placeName, TCL_VOLATILE);
    runUpdateCmds(interp);
  } else {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
       " placeName ?{lat lon}?", (char *)NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
updateCmd(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   *  Add or delete a command to execute if geoPlace database changes.
   *  Used to update widgets when a geoplace is created, destroyed, or moved.
   *  Usage:
   *    geoplace updatecmd add|rm id ?text?
   *      add means add a command to the list
   *      rm means remove it
   *      id is identifier.  Used for later reference.
   *      text = string sent to Tcl_Eval
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   */
  
  char
    *id,		/* Identifier for the command */
    *cmd, *newCmd;	/* Text of the command */
  int newPtr;
  Tcl_HashEntry *entry;

  if (argc == 5 && strcmp(argv[2], "add") == 0) {

    /* Add an update command */
    id = argv[3];
    cmd = argv[4];
    entry = Tcl_CreateHashEntry(&updateCmds, id, &newPtr);
    if ( !newPtr ) {
      char *oldCmd = (char *)Tcl_GetHashValue(entry);
      free(oldCmd);
    }
    if ( !(newCmd = (char *)malloc((strlen(cmd) + 1) * sizeof(char))) ) {
      Tcl_AppendResult(interp, argv[0], " could not allocate command string",
                       NULL);
      return TCL_ERROR;
    }
    strcpy(newCmd, cmd);
    Tcl_SetHashValue(entry, newCmd);

  } else if (argc == 4 && strcmp(argv[2], "rm") == 0) {

    /* Remove an update command */
    id = argv[3];
    if ( !(entry = Tcl_FindHashEntry(&updateCmds, id)) ) {
      Tcl_AppendResult(interp, argv[0], ": no update command named ", id, NULL);
      return TCL_ERROR;
    }
    cmd = (char *)Tcl_GetHashValue(entry);
    free(cmd);
    Tcl_DeleteHashEntry(entry);

  } else if (argc == 3) {

    /* Set result to text of an update command */
    id = argv[2];
    if ( !(entry = Tcl_FindHashEntry(&updateCmds, id)) ) {
      Tcl_AppendResult(interp, argv[0], ": no update command named ", id, NULL);
      return TCL_ERROR;
    }
    cmd = (char *)Tcl_GetHashValue(entry);
    Tcl_SetResult(interp, cmd, TCL_VOLATILE);

  } else {

    /* Bad call */
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " updatecmd add|rm id "
                     "?text?\"", NULL);
    return TCL_ERROR;

  }

  return TCL_OK;
}

void runUpdateCmds(Tcl_Interp *interp)
{
  /* Run the commands stored in updateCmds */
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;
  for (entry = Tcl_FirstHashEntry(&updateCmds, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    char *script = Tcl_GetHashValue(entry);
    if (Tcl_Eval(interp, script) == TCL_ERROR) {
      Tcl_AppendResult(interp, "Could not run \"", script, "\"", NULL);
    }
  }
}

int
rmPlace(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* rmPlace - remove a place of interest
   * Usage:
   *   geoplace rm placeName
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Appends error message to interp->result on failure
   */

  Tcl_HashEntry *entry;		/* Entry for place */
  GeoPlace *place;		/* GeoPlace to remove */
  char *placeName;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1], " placeName ",
		       (char *)NULL);
    return TCL_ERROR;
  }
  placeName = argv[2];
  if ((entry = Tcl_FindHashEntry(&places, placeName)) != NULL) {
    place = Tcl_GetHashValue(entry);
    if (place->userCnt == 0) {
      free(place);
      Tcl_DeleteHashEntry(entry);
    } else {
      Tcl_AppendResult(interp, placeName, " in use", (char *)NULL);
      return TCL_OK;
    }
  }
  return TCL_OK;
}

int
listPlaces(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* listPlaces - list all places
   * Usage:
   *   geoplace list
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Appends list of all known places to interp->result
   */

  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  for (entry = Tcl_FirstHashEntry(&places, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    Tcl_AppendElement(interp, Tcl_GetHashKey(&places, entry));
  }
  return TCL_OK;
}

int
distance(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* distance - find distance between two places or two coordinate pairs
   * Usage:
   *   geoplace distance place1|coords1 place2|coords2 ?unit?
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets interp->result to distance between two points, in great circle
   *     degrees, or appends error message to interp->result on failure
   */
  
  Tcl_HashEntry *entry;
  GeoPlace *place1, *place2;
  GeoPoint latLon1, latLon2;

  if (argc < 4 || argc > 5) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		       " placeOR{lat lon} placeOR{lat lon}"
		       "?nmi|nMi|smi|sMi|km|gsdeg|gsDeg?", (char *)NULL);
    return TCL_ERROR;
  }

  /* Identify the first place */
  if ((entry = Tcl_FindHashEntry(&places, argv[2])) != NULL) {
    place1 = Tcl_GetHashValue(entry);
    latLon1 = place1->latLon;
  } else if (sscanf(argv[2], "%f %f", &latLon1.ord, &latLon1.abs) == 2) {
    /* argv[2] is a {lat lon} */
  } else {
    Tcl_AppendResult(interp, argv[2], " not a location", (char *)NULL);
    return TCL_ERROR;
  }

  /* Identify the second place */
  if ((entry = Tcl_FindHashEntry(&places, argv[3])) != NULL) {
    place2 = Tcl_GetHashValue(entry);
    latLon2 = place2->latLon;
  } else if (sscanf(argv[3], "%f %f", &latLon2.ord, &latLon2.abs) == 2) {
    /* argv[3] is a {lat lon} */
  } else {
    Tcl_AppendResult(interp, argv[3], " not a location", (char *)NULL);
    return TCL_ERROR;
  }

  /* Identify optional units.  Otherwise result is in great circle degrees */
  if (argc == 5) {
    /* User specified distance unit */
    char *unit = argv[4];

    if ( strcmp(unit, "nmi") == 0 || strcmp(unit, "nMi") == 0 ) {
      sprintf(result, "%7.2f", GeoDistance(latLon1, latLon2) * NMIPERDEG);
    } else if ( strcmp(unit, "smi") == 0 || strcmp(unit, "sMi") == 0 ) {
      sprintf(result, "%7.2f", GeoDistance(latLon1, latLon2) * SMIPERDEG);
    } else if ( strcmp(unit, "km") == 0 ) {
      sprintf(result, "%7.2f", GeoDistance(latLon1, latLon2) * KMPERDEG);
    } else if ( strcmp(unit, "gsdeg") == 0 || strcmp(unit, "gsDeg") == 0 ) {
      sprintf(result, "%7.2f", GeoDistance(latLon1, latLon2));
    } else {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], " unknown unit: ",
                       unit, NULL);
      return TCL_ERROR;
    }
  } else {
    /* Default unit */
    sprintf(result, "%7.2f", GeoDistance(latLon1, latLon2));
  }

  /* Set result to distance between the two places */
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
nearest(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * nearest - find place nearest a given latitude and longitude
   *   Searches for nearest of all places or from optional list.
   * Usage:
   *   geoplace nearest placeORlatLon ?place place ...?
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets interp result to name of nearest place,
   *   or appends error message interp->result on failure
   */
  
  GeoPoint latLon;		/* Location for which nearest place  sought */
  GeoPlace *place;		/* Current place */
  char *skip = NULL;		/* Hold argv[2] if it is a placeName */
  double
    nearDist = 181.0,		/* Hold distance to nearest place so far */
    cDistance;			/* Distance to current place */
  Tcl_HashEntry
    *entry,			/* Entry for current place */
    *nearEntry = NULL;		/* Entry for nearest place found so far */
  Tcl_HashSearch search;

  /* Check the call */
  if (argc < 3 || argc > 4) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
      " place [place place ...]", (char *)NULL);
    return TCL_ERROR;
  }

  /* argv[2] is either a placeName or a {lat lon} */
  if (sscanf(argv[2], "%f %f", &latLon.ord, &latLon.abs) == 2) {
    /* Second argument is a latLon pair */
  } else if ( (entry = Tcl_FindHashEntry(&places, argv[2])) ) {
    place = Tcl_GetHashValue(entry);
    latLon = place->latLon;
    skip = argv[2];
  } else {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not interpret ",
       argv[2], " as a place.", (char *)NULL);
    return TCL_ERROR;
  }

  /* Find the nearest place */
  if (argc == 3) {
    /* Search entire geoplace database */
    for (entry = Tcl_FirstHashEntry(&places, &search);
         entry != NULL;
         entry = Tcl_NextHashEntry(&search)) {
      if (skip && (strcmp(skip, Tcl_GetHashKey(&places, entry)) == 0))
        continue;	/* Skip over our reference point */
      place = Tcl_GetHashValue(entry);
      cDistance = GeoQuickDistance(place->latLon, latLon);
  
      if (cDistance < nearDist) {
        nearDist = cDistance;
        nearEntry = entry;
      }
    }
  } else if (argc == 4) {
    /* Search list of placeNames */
    int np, placeCnt;
    char **placeList;

    if (Tcl_SplitList(interp, argv[3], &placeCnt, &placeList) != TCL_OK) {
      return TCL_ERROR;
    }
    for (np = 0; np < placeCnt; np++) {
      if ( !(entry = Tcl_FindHashEntry(&places, placeList[np])) ) {
        Tcl_AppendResult(interp, argv[0], " ", argv[1], ": no place named ",
	  placeList[np], (char *)NULL);
        free(placeList);
        return TCL_ERROR;
      }
      if (skip && (strcmp(skip, Tcl_GetHashKey(&places, entry)) == 0))
        continue;	/* Skip over our reference point */
      place = Tcl_GetHashValue(entry);
      cDistance = GeoQuickDistance(place->latLon, latLon);
      if (cDistance < nearDist) {
        nearDist = cDistance;
        nearEntry = entry;
      }
    }
    free(placeList);
  }

  Tcl_SetResult(interp, Tcl_GetHashKey(&places, nearEntry), TCL_STATIC);
  return TCL_OK;
}

int
step(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* step - find point at given bearing (degrees cw from north) and range (in
   *   Great Circle degrees) from given point
   * Usage:
   *   geoplace step place bearing range ?unit?
   *     place = placeName or {lat lon}
   *     bearing = direction in which to step (degrees from north)
   *     range = distance to step (Great Circle degrees)
   *     unit = optional unit nmi, nMi, smi, sMi, km, gsdeg, or gsDeg
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets interp result lat-lon of point at given bearing and range from point
   *   Appends error message to interp->result on failure
   */
  
  GeoPoint latLon;		/* Starting location */
  GeoPoint endPt;		/* Point at bearing and range from latLon */
  GeoPlace *place;		/* Current place */
  double brg, rng;		/* Desired bearing and range */
  Tcl_HashEntry *entry;		/* Entry for current place */

  if (argc != 5 && argc != 6) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " step place bearing range "
		       "?nmi|nMi|smi|sMi|km|gsdeg|gsDeg?", NULL);
    return TCL_ERROR;
  }

  /* Read third argument as place name or coordinate pair */
  if ((entry = Tcl_FindHashEntry(&places, argv[2]))) {
    place = Tcl_GetHashValue(entry);
    latLon = place->latLon;
  } else if (sscanf(argv[2], "%f %f", &latLon.ord, &latLon.abs) == 2) {
    /* arg is a latLon pair */
  } else {
    Tcl_AppendResult(interp, argv[0], " could not read place ", argv[2],
       (char *)NULL);
    return TCL_ERROR;
  }

  /* Read bearing */
  if (sscanf(argv[3], "%lf", &brg) != 1) {
    Tcl_AppendResult(interp, argv[0], " could not read bearing ", argv[3],
       (char *)NULL);
    return TCL_ERROR;
  }

  /* Read range */
  if (sscanf(argv[4], "%lf", &rng) != 1) {
    Tcl_AppendResult(interp, argv[0], " could not read range ", argv[4],
       (char *)NULL);
    return TCL_ERROR;
  }

  /* Adjust distance unit */
  if (argc == 6) {
    char *unit = argv[5];

    if ( strcmp(unit, "nmi") == 0 || strcmp(unit, "nMi") == 0 ) {
      rng /= NMIPERDEG;
    } else if ( strcmp(unit, "smi") == 0 || strcmp(unit, "sMi") == 0 ) {
      rng /= SMIPERDEG;
    } else if ( strcmp(unit, "km") == 0 ) {
      rng /= KMPERDEG;
    } else if ( strcmp(unit, "gsdeg") == 0 || strcmp(unit, "gsDeg") == 0 ) {
      /* No conversion required */
    } else {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], " unknown unit: ",
                       unit, NULL);
      return TCL_ERROR;
    }
  }

  endPt = GeoStep(latLon, brg, rng);
  sprintf(result, "%g %g", endPt.ord, endPt.abs);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
inrange(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* inrange - find places within given distance of given point
   * Usage:
   *   geoplace inrange {lat lon}ORplace {range ?unit?} [place place ...]
   *     place = name of place or {lat lon} we are measuring distances from.
   *     range = distance (Great Circle degrees) we are going out. 
   *     [place place ...] = optional list of places to evaluate.
   *       If place list absent, inrange searches entire geoplace database.
   *     unit  = optional distance unit
   * Input:
   *   clientData - not used
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Return value:
   *   TCL_OK or TCL_ERROR
   * Side effects:
   *   Sets interp result to list of places within given range of place
   *   Appends error message to interp->result on failure
   */

  GeoPoint latLon;		/* Location we are measuring from */
  GeoPoint
    refPt,			/* References to use in units conversion */
    origin = {0.0, 0.0};
  double rng;			/* Look for places within range of latLon */
  GeoPlace *place;		/* Current place */
  Tcl_HashEntry *entry;		/* Entry for current place */
  Tcl_HashSearch search;
  char unit[8];

  /* Check the call */
  if (argc < 4 || argc > 5) {
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " ", argv[1],
      " placenameOR{lat lon} range ?place place place ...?\" or \""
      "\"", argv[0], " ", argv[1], " placenameOR{lat lon} {range unit} "
      " ?place place place ...?\"", (char *)NULL);
    return TCL_ERROR;
  }

  /* argv[2] is either a placeName or a {lat lon} */
  if (sscanf(argv[2], "%f %f", &latLon.ord, &latLon.abs) == 2) {
    /* Second argument is a latLon pair */
  } else if ( (entry = Tcl_FindHashEntry(&places, argv[2])) ) {
    place = Tcl_GetHashValue(entry);
    latLon = place->latLon;
  } else {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not interpret ",
       argv[2], " as a place.", (char *)NULL);
    return TCL_ERROR;
  }

  /* Get range */
  if (sscanf(argv[3], "%lf %s", &rng, unit) == 2) {

    if ( strcmp(unit, "nmi") == 0 || strcmp(unit, "nMi") == 0 ) {
      rng /= NMIPERDEG;
    } else if ( strcmp(unit, "smi") == 0 || strcmp(unit, "sMi") == 0 ) {
      rng /= SMIPERDEG;
    } else if ( strcmp(unit, "km") == 0 ) {
      rng /= KMPERDEG;
    } else if ( strcmp(unit, "gsdeg") == 0 || strcmp(unit, "gsDeg") == 0 ) {
      /* No conversion */
    } else {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], " unknown unit: ",
                       unit, NULL);
      return TCL_ERROR;
    }

  } else if (sscanf(argv[3], "%lf", &rng) == 1) {

    /* No conversion */

  } else {

    Tcl_AppendResult(interp, argv[0], argv[1], " could not interpret ",
      argv[3], " as a range", (char *)NULL);
    return TCL_ERROR;

  }

  /* Make a fictitious point {rng 0} and use to convert range to units
   * returned by GeoQuickDistance (i.e. square of absolute distance) by
   * measuring from {0 0} to {rng 0} */
  refPt.ord = rng;
  refPt.abs = 0.0;
  rng = GeoQuickDistance(refPt, origin);

  /* Find places within range of latLon */
  if (argc == 4) {
    /* Search entire geoplace database */
    for (entry = Tcl_FirstHashEntry(&places, &search);
         entry != NULL;
         entry = Tcl_NextHashEntry(&search)) {
      place = Tcl_GetHashValue(entry);
      if (GeoQuickDistance(place->latLon, latLon) < rng) {
        Tcl_AppendElement(interp, Tcl_GetHashKey(&places, entry));
      }
    }
  } else if (argc == 5) {
    /* Search list of placeNames */
    int np, placeCnt;
    char **placeList;

    if (Tcl_SplitList(interp, argv[4], &placeCnt, &placeList) != TCL_OK) {
      return TCL_ERROR;
    }
    for (np = 0; np < placeCnt; np++) {
      if ( !(entry = Tcl_FindHashEntry(&places, placeList[np])) ) {
        Tcl_AppendResult(interp, argv[0], " ", argv[1], ": no place named ",
	  placeList[np], (char *)NULL);
        free(placeList);
        return TCL_ERROR;
      }
      place = Tcl_GetHashValue(entry);
      if (GeoQuickDistance(place->latLon, latLon) < rng) {
        Tcl_AppendElement(interp, Tcl_GetHashKey(&places, entry));
      }
    }
    free(placeList);
  }

  return TCL_OK;
}

int
lonbtwn(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * Set result to "1" if a lon is between two other lons, else "0"
   * Usage:
   *   geoplace lonbtwn lon {lon0 lon1}
   */

  double lon, lon0, lon1;

  if (   argc != 4
      || sscanf(argv[2], "%lf", &lon) != 1
      || sscanf(argv[3], "%lf %lf", &lon0, &lon1) != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " lonbtwn lon {lon0 lon1}",
      NULL);
    return TCL_ERROR;
  }
  
  Tcl_SetResult(interp, LonBtwn(lon, lon0, lon1) ? "1" : "0", TCL_STATIC);
  return TCL_OK;
  
}
