/*
 * tkGeoMap - a tk widget to draw geographic maps.
 *
 * Copyright (C) 1999 Gordon D. Carrie
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * 
 * Please address questions and bug reports to gcarri@mail.txcyber.com
 *
 **********************************************************************
 *
 */
 
/* 
 * Based on tkSquare.c.  See the file tkSquare.c for copywrite and disclaimers.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <tcl.h>
#include <tk.h>
#include "tkGeoMap.h"
#include "tkConvertD.h"
#include "geography.h"
#include "geoProj.h"
#include "geoLines.h"
#include "tclGeoLines.h"
#include "tclGeoPlace.h"

/* Struct defining parts of a geomap widget */
typedef struct {

  /******** Widget ********/

  Tk_Window     tkwin;          /* Window that contains the geomap.
                                 * NULL means has been deleted but widget
                                 * record hasn't been cleaned up yet. 
                                 */

  char		*name;		/* Pathname */

  Display       *display;       /* X's token for the window's display. */

  Tcl_Interp    *interp;        /* Interpreter associated with widget. */

  GC            gc;             /* Graphics context */

  int           borderWidth;    /* Width of 3-D border around whole widget. */

  Tk_3DBorder   bgBorder;       /* Used for drawing background. */

  int           width;          /* Window width */

  int           height;         /* Window height */

  int           relief;         /* Whole window is raised, sunken, or flat. */

  Cursor        cursor;         /* Widget cursor */

  Tcl_HashTable lines;          /* Table of arrays of geographical outlines */

  Tcl_HashTable places;         /* Table of places of interest */

  int           updatePending;  /* Non-zero => call geoMapXDisplay when idle */

  int		drawPending;	/* Non-zero => call geoMapDraw when idle */

  /******** Map ********/

  GeoPoint      center;         /* LatLon of visible-window center */
        
  Tk_Uid        scale;		/* Map scale as a character string */

  Projection    projection;     /* Projection information */

  /******** Lines ********/

  XColor        *lineColor;     /* Color for drawing outlines */

  /******** Bitmaps ********/

  XColor        *bmColor;       /* Default bitmap color */

  /******** Dots ********/

  int           dotSize;        /* Default dot size */

  XColor        *dotColor;      /* Default dot color */

  /******** Text ********/

  Tk_Uid	fontName;	/* Name of default font */

  XFontStruct*	font;		/* Default font */

  XColor        *textColor;     /* Default text color */

  Tk_Anchor     textAnchor;     /* Location of text relative to point */

} GeoMap;

/* Struct for an array of lines to draw */
typedef struct {
  GeoLineArray	projLineArray;	/* Projection coord's of points in the lines */
  XColor	*lineColor;	/* Line color */
  int		lineWidth;	/* Line width, in pixels */
  int		lineStyle;	/* LineSolid, LineOnOffDash,  LineDoubleDash */
  int		dotSize;	/* Size of dots at segment ends */
  Pixmap	pixmap;		/* Pixmap to draw at segment ends */
} GeoMapLineArray;

/* Struct defining a place of interest */
typedef struct {
  GeoPoint      latLon;		/* Lat-lon of the place */
  GeoPoint      projPoint;	/* Projection coordinates of the place */
  XColor        *dotColor;	/* Dot color. NULL => use widget default fg */
  int           dotSize;	/* Dot size.  0 => no dot, <0 => use default */
  Pixmap        pixmap;		/* Pixmap to draw */
  XColor        *bmColor;	/* Bitmap color NULL => use widget  fg */
  char          *text;		/* Associated text information */
  XColor        *textColor;	/* Text color. NULL => use default fg */
  XFontStruct   *font;		/* Text font. NULL => use default font */
  Tk_Anchor     *textAnchor;	/* Relative location of text. */
} GeoMapPlace;

/* Tell if package already loaded */
static int loaded;

/* Place to write a formatted result */
static char result[200];

/* Local functions */
static int
initCmdTable(Tcl_Interp *interp);
static Tcl_CmdProc commands;
static Tcl_CmdProc configure;
static Tcl_CmdProc cget;
static Tcl_CmdProc update;
static Tcl_CmdProc center;
static Tcl_CmdProc projection;
static Tcl_CmdProc rotation;
static Tcl_CmdProc latLonToProj;
static Tcl_CmdProc xToLatLon;
static Tcl_CmdProc latLonToX;
static Tcl_CmdProc projToLatLon;
static Tcl_CmdProc shift;
static Tcl_CmdProc drawLines;
static Tcl_CmdProc eraseLines;
static Tcl_CmdProc drawBitmap;
static Tcl_CmdProc eraseBitmap;
static Tcl_CmdProc drawDot;
static Tcl_CmdProc eraseDot;
static Tcl_CmdProc text;
static Tcl_CmdProc eraseText;
static Tcl_CmdProc erase;
static Tcl_CmdProc listWdgPlaces;
static Tcl_CmdProc geoMapCmd;
static Tcl_CmdProc geoMapWidgetCmd;

static int
geoMapConfigure(Tcl_Interp *interp, GeoMap *geoMapPtr, int argc, char **argv,
		int flags);
static void
geoMapEventProc(ClientData clientData, XEvent *eventPtr);

static void
geoMapDraw(ClientData clientData);

static void
geoMapXDisplayWhenIdle(GeoMap *geoMapPtr);

static void
geoMapXDisplay(ClientData clientData);

static Tcl_FreeProc geoMapDestroy;

static GeoMapPlace *
geoMapCreatePlace(GeoMap *geoMapPtr, char *placeName);

static void
geoMapRmPlace(GeoMap *geoMapPtr, char *placeName);

static void
shiftArea(Display *display, Drawable mvPic, GC gc, int x, int y,
	  unsigned width, unsigned height, int deltaX, int deltaY,
	  unsigned depth);

/************************ End of declarations *************************/

/* Table of subcommands for a geomap widget */
static Tcl_HashTable cmdTbl;

/* Widget configuration options */
static Tk_ConfigSpec configSpecs[] = {
  
  {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth", "4",
     Tk_Offset(GeoMap, borderWidth), 0, NULL},

  {TK_CONFIG_SYNONYM, "-bd", "borderWidth", NULL, NULL, 0, 0, NULL},

  {TK_CONFIG_BORDER, "-background", "background", "Background", "#cdb79e",
     Tk_Offset(GeoMap, bgBorder), TK_CONFIG_COLOR_ONLY, NULL},

  {TK_CONFIG_BORDER, "-background", "background", "Background", "white",
     Tk_Offset(GeoMap, bgBorder), TK_CONFIG_MONO_ONLY, NULL},

  {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0, NULL},

  {TK_CONFIG_INT, "-width", "width", "Width", "400", Tk_Offset(GeoMap, width),
     0, NULL},

  {TK_CONFIG_INT, "-height", "height", "Height", "400",
     Tk_Offset(GeoMap, height), 0, NULL},

  {TK_CONFIG_RELIEF, "-relief", "relief", "Relief", "sunken",
     Tk_Offset(GeoMap, relief), 0, NULL},

  {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
     "top_left_arrow", Tk_Offset(GeoMap, cursor), 0, NULL},

  {TK_CONFIG_UID, "-scale", "scale", "Scale", "1:20000000",
     Tk_Offset(GeoMap, scale), 0, NULL},

  {TK_CONFIG_COLOR, "-linecolor", "linecolor", "Foreground", "NavyBlue",
     Tk_Offset(GeoMap, lineColor), TK_CONFIG_COLOR_ONLY, NULL},

  {TK_CONFIG_COLOR, "-linecolor", "linecolor", "Foreground", "black",
     Tk_Offset(GeoMap, lineColor), TK_CONFIG_MONO_ONLY, NULL},

  {TK_CONFIG_SYNONYM, "-lc", "linecolor",  NULL, NULL, 0, 0, NULL},

  {TK_CONFIG_INT, "-dotsize", "dotsize", "Dotsize", "1",
     Tk_Offset(GeoMap, dotSize), 0, NULL},

  {TK_CONFIG_COLOR, "-dotcolor", "dotcolor", "Foreground", "NavyBlue",
     Tk_Offset(GeoMap, dotColor), TK_CONFIG_COLOR_ONLY, NULL},

  {TK_CONFIG_COLOR, "-dotcolor", "dotcolor", "Foreground", "black",
     Tk_Offset(GeoMap, dotColor), TK_CONFIG_MONO_ONLY, NULL},
    
  {TK_CONFIG_SYNONYM, "-dc", "dotcolor",  NULL, NULL, 0, 0, NULL},
    
  {TK_CONFIG_COLOR, "-bmColor", "bmColor", "Foreground", "Brown4",
     Tk_Offset(GeoMap, bmColor), TK_CONFIG_COLOR_ONLY, NULL},
    
  {TK_CONFIG_COLOR, "-bmcolor", "bmcolor", "Foreground", "black",
     Tk_Offset(GeoMap, bmColor), TK_CONFIG_MONO_ONLY, NULL},
    
  {TK_CONFIG_SYNONYM, "-bc", "bmcolor",  NULL, NULL, 0, 0, NULL},
    
  {TK_CONFIG_UID, "-fn", "font", "Font", "fixed", Tk_Offset(GeoMap, fontName),
     0, NULL},
    
  {TK_CONFIG_SYNONYM, "-font", "font",  NULL, NULL, 0, 0, NULL},
    
  {TK_CONFIG_COLOR, "-textcolor", "textcolor", "Foreground", "NavyBlue",
     Tk_Offset(GeoMap, textColor), TK_CONFIG_COLOR_ONLY, NULL},
    
  {TK_CONFIG_COLOR, "-textcolor", "textcolor", "Foreground", "black",
     Tk_Offset(GeoMap, textColor), TK_CONFIG_MONO_ONLY, NULL},
    
  {TK_CONFIG_SYNONYM, "-tc", "textcolor",  NULL, NULL, 0, 0, NULL},
    
  {TK_CONFIG_ANCHOR, "-anchor", "anchor", "Anchor", "ne",
     Tk_Offset(GeoMap, textAnchor), 0, NULL},
    
  {TK_CONFIG_END,  NULL,  NULL,  NULL, NULL, 0, 0, NULL}
};

int
Tkgeomap_Init(Tcl_Interp *interp)
{
  /* GeoMapInit -- creates the geomap command in interp, creates table of
   * geomap subcommands, and initializes auxilliary packages
   * Input: interp - tcl interpreter
   */

  Tk_Window mainWin = Tk_MainWindow(interp);

  if (loaded)
    return TCL_OK;
  loaded = 1;

  /* Create widget command */
  Tcl_CreateCommand(interp, "tkgeomap", geoMapCmd, (ClientData)mainWin,
                    (void (*)())NULL);
  Tcl_CreateCommand(interp, "geomap", geoMapCmd, (ClientData)mainWin,
                    (void (*)())NULL);

  /* Initialize table of widget subcommands */
  if (initCmdTable(interp) == TCL_ERROR) {
    Tcl_AppendResult(interp, "Could not load commands.", NULL);
    return TCL_ERROR;
  }

  /* Init required packages */
  Tclgeolines_Init(interp);
  Tclgeoplace_Init(interp);
  Tkconvertd_Init(interp);

  /* Provide this as a package */
  Tcl_PkgProvide(interp, "tkgeomap", "1.2");

  return TCL_OK;
}

int
initCmdTable(Tcl_Interp *interp)
{
  /* initCmdTable - Create table to map geomap widget subcommand names to
   * their functions
   * Input: interp - tcl interpreter
   */

  /* Struct to map command name to command function */
  struct NamedCommand {
    char	*name;
    Tcl_CmdProc	*function;
  };
  
  /* Commands for GeoMap widget */
  struct NamedCommand commandList[] = {
    {"bitmap",		drawBitmap	},
    {"Center",		center		},
    {"center",		center		},
    {"cget",		cget		},
    {"Commands",	commands	},
    {"commands",	commands	},
    {"config",		configure	},
    {"configure",	configure	},
    {"dot",		drawDot		},
    {"DrawBitmap",	drawBitmap	},
    {"drawbitmap",	drawBitmap	},
    {"DrawDot",		drawDot		},
    {"drawdot",		drawDot		},
    {"DrawLines",	drawLines	},
    {"drawlines",	drawLines	},
    {"Erase",		erase		},
    {"erase",		erase		},
    {"EraseBitmap",	eraseBitmap	},
    {"erasebitmap",	eraseBitmap	},
    {"EraseDot",	eraseDot	},
    {"erasedot",	eraseDot	},
    {"EraseLines",	eraseLines	},
    {"eraselines",	eraseLines	},
    {"EraseText",	eraseText	},
    {"erasetext",	eraseText	},
    {"LatLonToProj",	latLonToProj	},
    {"latlontoproj",	latLonToProj	},
    {"LatLonToX",	latLonToX	},
    {"latlontox",	latLonToX	},
    {"lines",		drawLines	},
    {"ListPlaces",	listWdgPlaces	},
    {"listplaces",	listWdgPlaces	},
    {"Proj",		projection	},
    {"proj",		projection	},
    {"Projection",	projection	},
    {"projection",	projection	},
    {"ProjToLatLon",	projToLatLon	},
    {"projtolatlon",	projToLatLon	},
    {"Rot",		rotation	},
    {"rot",		rotation	},
    {"Rotation",	rotation	},
    {"rotation",	rotation	},
    {"Shift",		shift		},
    {"shift",		shift		},
    {"Text",		text		},
    {"text",		text		},
    {"update",		update		},
    {"XToLatLon",	xToLatLon	},
    {"xtolatlon",	xToLatLon	},
    {NULL,		NULL		}
  };
  struct NamedCommand *command;		/* Loop index */
  int np;            			/* Ignored */

  Tcl_InitHashTable(&cmdTbl, TCL_STRING_KEYS);

  /* Add commands to GeoMap command table */
  for (command = commandList; command->name; command++)
    Tcl_SetHashValue(Tcl_CreateHashEntry(&cmdTbl, command->name, &np),
                     (ClientData)command->function);
  return TCL_OK;
}

int
geoMapCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
  /* geoMapCmd -- Called when geomap command is called in Tcl
   * Usage:
   *   geomap pathName ?options?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Side effects:
   *   Creates a new widget
   *   Sets interp->result to pathName of new widget or error message
   */
  
  Tk_Window mainWin = (Tk_Window)clientData;
  GeoMap *geoMapPtr;
  Tk_Window tkwin;
  char *cmd;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # args:  should be \"", argv[0],
                     " pathName ?options?\"", NULL);
    return TCL_ERROR;
  }

  if ( !(tkwin = Tk_CreateWindowFromPath(interp, mainWin, argv[1], NULL)) )
    return TCL_ERROR;

  Tk_SetClass(tkwin, "TkGeoMap");

  /* Allocate and initialize the widget record. */
  if ( !(geoMapPtr = calloc(1, sizeof(GeoMap))) ) {
    Tcl_AppendResult(interp, "Could not allocate storage for map", NULL);
    return TCL_ERROR;
  }
        
  /* Initialize non-null members */
  geoMapPtr->tkwin = tkwin;
  geoMapPtr->display = Tk_Display(tkwin);
  geoMapPtr->interp = interp;
  geoMapPtr->gc = None;

  if ( !(geoMapPtr->name = (char *)malloc((strlen(argv[1])+1)*sizeof(char))) ) {
    Tcl_SetResult(interp, "Could not save name", TCL_STATIC);
    free(geoMapPtr);
    return TCL_ERROR;
  }
  strcpy(geoMapPtr->name, argv[1]);
  
  /* Initialize widget tables */
  Tcl_InitHashTable(&geoMapPtr->lines, TCL_STRING_KEYS);
  Tcl_InitHashTable(&geoMapPtr->places, TCL_STRING_KEYS);

  /* Initialize pointers */
  geoMapPtr->lineColor = NULL;
  geoMapPtr->bmColor = NULL;
  geoMapPtr->dotColor = NULL;
  geoMapPtr->font = NULL;
  geoMapPtr->textColor = NULL;

  /* Default center */
  geoMapPtr->center.ord = 0.0;
  geoMapPtr->center.abs = 0.0;

  /* Default to cylindrical equidistant projection */
  if ( !(geoMapPtr->projection = AllocProj())) {
    Tcl_AppendResult(interp, argv[0], " could not create projection");
    free(geoMapPtr);
    return TCL_ERROR;
  }
  SetCylEqdist(geoMapPtr->projection, 0.0);

  Tk_CreateEventHandler(geoMapPtr->tkwin, ExposureMask|StructureNotifyMask,
                        geoMapEventProc, (ClientData) geoMapPtr);
  Tcl_CreateCommand(interp, geoMapPtr->name, geoMapWidgetCmd,
                    (ClientData)geoMapPtr, (void (*)())NULL);

  /* Request update when a geoplace changes */
  if ( !(cmd = (char *)malloc((strlen(geoMapPtr->name) + strlen(" update") + 1)
                           * sizeof(char))) ) {
    Tcl_AppendResult(interp, "Could not create update command for ",
      geoMapPtr->name, NULL);
    free(geoMapPtr);
    return TCL_ERROR;
  }
  strcpy(cmd, geoMapPtr->name);
  strcat(cmd, " update");
  if (Tcl_VarEval(interp, "geoplace updatecmd add ", geoMapPtr->name, " {",
                  cmd, "}", NULL) == TCL_ERROR) {
    Tcl_AppendResult(interp, "Could not add update command for ",
                     geoMapPtr->name, NULL);
    return TCL_ERROR;
  }

  if (geoMapConfigure(interp, geoMapPtr, argc-2, argv+2, 0) != TCL_OK) {
    Tcl_AppendResult(interp, "Could not configure new geomap.", NULL);
    Tk_DestroyWindow(geoMapPtr->tkwin);
    Tcl_VarEval(interp, "geoplace updatecmd rm ", geoMapPtr->name, " {",
                  cmd, "}", NULL);
    free(cmd);
    free(geoMapPtr);
    return TCL_ERROR;
  }
  Tcl_SetResult(interp, geoMapPtr->name, TCL_VOLATILE);
  return TCL_OK;
}

int
geoMapWidgetCmd(ClientData clientData, Tcl_Interp *interp, int argc,
		char **argv)
{
  /* geoMapWidgetCmd - call function corresponding to a widget command
   * Usage:
   *   pathName commandName ?arg arg ...?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;
  int result;
  Tcl_HashEntry *cmdEntry;
  Tcl_CmdProc *command;

  if (argc < 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
                     " option ?arg arg ...?\"", NULL);
    return TCL_ERROR;
  }
  Tk_Preserve((ClientData) geoMapPtr);

  /* Get the command from the geomap widget command table and execute it */
  if ( (cmdEntry = Tcl_FindHashEntry(&cmdTbl, argv[1])) ) {
    command = (Tcl_CmdProc *)Tcl_GetHashValue(cmdEntry);
    result = command(geoMapPtr, interp, argc, argv);
  } else {
    Tcl_AppendResult(interp, argv[0], "unknown command: ", argv[1], NULL);
    result = TCL_ERROR;
  }
  Tk_Release((ClientData) geoMapPtr);
  return result;
}

int
commands(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * Set interp->result to a list of commands for a GeoMap widget
   * Usage:
   *   pathName commands
   */
  
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  if (argc > 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " commands", NULL);
    return TCL_ERROR;
  }

  for (entry = Tcl_FirstHashEntry(&cmdTbl, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search))
    Tcl_AppendResult(interp, Tcl_GetHashKey(&cmdTbl, entry), " ", NULL);
  return TCL_OK;
}

int
configure(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* configure - Configure widget or report on current options and values
   * Usage:
   *   pathName configure ?option? ?value? ?option value ...?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Side effects:
   *   Sets interp->result to config info or error message
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  int rtn;

  if (argc == 2) {
    rtn = Tk_ConfigureInfo(interp, geoMapPtr->tkwin, configSpecs,
                              (char *)geoMapPtr, NULL, 0);
  } else if (argc == 3) {
    rtn = Tk_ConfigureInfo(interp, geoMapPtr->tkwin, configSpecs,
                                (char *)geoMapPtr, argv[2], 0);
  } else {
    rtn = geoMapConfigure(interp, geoMapPtr, argc-2, argv+2,
                             TK_CONFIG_ARGV_ONLY);
  }
  return rtn;
}

int
cget(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* cget - implement "cget" command
   * Usage:
   *   pathName cget option
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   * Side effects:
   *   Sets interp->result to config info or error message
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " cget option", NULL);
    return TCL_ERROR;
  }
  return Tk_ConfigureValue(interp, geoMapPtr->tkwin, configSpecs,
                           (char *)geoMapPtr, argv[2], 0);
}

int
update(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * Request draw (recomputing projections, etc.) and display of geoMapPtr
   * Usage:
   *   pathName update [now]
   * Now means do NOT wait until tk is idle.
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc == 2) {
    if (   geoMapPtr->tkwin && Tk_IsMapped(geoMapPtr->tkwin)
        && !geoMapPtr->drawPending) {
      Tk_DoWhenIdle(geoMapDraw, (ClientData)geoMapPtr);
      geoMapPtr->drawPending = 1;
    }
  } else if (argc == 3 && strcmp(argv[2], "now") == 0) {
    geoMapDraw(geoMapPtr);
  } else {
    Tcl_AppendResult(interp, "Usage: \"", argv[0], " update ?now?\"", NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

int
center(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * Set result to or set lat and lon of window center.
   * Input and output are lat and lon.
   * Usage:
   *   pathName center placeOR{lat lon}
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc == 2) {
    sprintf(result, "%.3f %.3f",
            geoMapPtr->center.ord, geoMapPtr->center.abs);
    Tcl_SetResult(interp, result, TCL_VOLATILE);
    return TCL_OK;
  } else if (argc == 3) {
    if (   sscanf(argv[2], "%f %f",
                  &geoMapPtr->center.ord, &geoMapPtr->center.abs) == 2
        || GeoPlaceGet(interp, argv[2], &geoMapPtr->center, USE_PLACE_ONCE)
	   == TCL_OK) {
      geoMapPtr->center = GwchLonPt(geoMapPtr->center);
      geoMapXDisplayWhenIdle(geoMapPtr);
      return TCL_OK;
    } else {
      Tcl_AppendResult(interp, "Usage: \"", argv[0],
                       " center ?place?\" or \"", argv[0],
                       " center ?{lat lon}?\"", NULL);
      return TCL_ERROR;
    }
  } else {
    Tcl_AppendResult(interp, "Usage: \"", argv[0],
                     " center ?place?\" or \"", argv[0],
                     "center ?{lat lon}?\"", NULL);
    return TCL_ERROR;
  }
}

int
projection(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Print or set widget projection
   * Usage:
   *   pathName projection ?projectionName? ?opt opt ...?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap	*geoMapPtr = (GeoMap *)clientData;
  Projection	proj = geoMapPtr->projection;
  GeoPoint	refPoint;		/* Reference point */
  double	refLon;			/* Reference longitude */
      
  if (argc == 2) {
    /* Set result to projection information */
    Tcl_SetResult(interp, ProjDescriptor(proj), TCL_VOLATILE);
    return TCL_OK;
  }
        
  if (   strcmp("CylEqdist", argv[2]) == 0
      || strcmp("CylEqDist", argv[2]) == 0
      || strcmp("cyleqdist", argv[2]) == 0) {
      
    if (argc == 3) {
      /* Use longitude of widget center as reference longitude */
      SetCylEqdist(geoMapPtr->projection, geoMapPtr->center.abs);
    } else if (argc == 4 && sscanf(argv[3], "%lf", &refLon) == 1) {
      /* Use given reference longitude */
      SetCylEqdist(geoMapPtr->projection, refLon);
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], "projection CylEqdist "
                       "?reflon?", NULL);
      return TCL_ERROR;
    }

  } else if (   strcmp("Mercator", argv[2]) == 0
             || strcmp("mercator", argv[2]) == 0) {
      
    if (argc == 3) {
      /* Use longitude of widget center as reference longitude */
      SetMercator(geoMapPtr->projection, geoMapPtr->center.abs);
    } else if (argc == 4 && sscanf(argv[3], "%lf", &refLon) == 1) {
      /* Use given reference longitude */
      SetMercator(geoMapPtr->projection, refLon);
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], "projection Mercator "
                       "[reflon|{lat, reflon}]", NULL);
      return TCL_ERROR;
    }

  } else if (   strcmp("CylEqArea", argv[2]) == 0
             || strcmp("cyleqarea", argv[2]) == 0
	     || strcmp("Cylindrical Equal Area", argv[2]) == 0
	     || strcmp("cylindrical equal area", argv[2]) == 0) {
      
    if (argc == 3) {
      /* Use longitude of widget center as reference longitude */
      SetCylEqArea(geoMapPtr->projection, geoMapPtr->center.abs);
    } else if (argc == 4 && sscanf(argv[3], "%lf", &refLon) == 1) {
      /* Use given reference longitude */
      SetCylEqArea(geoMapPtr->projection, refLon);
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], "projection CylEqArea "
                       "?reflon?", NULL);
      return TCL_ERROR;
    }

  } else if (   strcmp("LambertConfConic", argv[2]) == 0
             || strcmp("LCC", argv[2]) == 0) {

    /* For Lambert conformal conic projection, user can give optional
     * third argument as either a named place or lat-lon pair.
     * This provides the reference latitude for the projection and says
     * which meridian should be vertical.  If no third argument, then
     * current map center is the reference point (default)
     */
    double rlat, rlon;	/* Components of reference point */

    if (argc == 3) {
      /* Change projection without moving center */
      refPoint = geoMapPtr->center;
    } else if (sscanf(argv[3], "%lf %lf", &rlat, &rlon) == 2) {
      /* Reference point is a lat-lon pair.  No further action */
    } else if (GeoPlaceGet(interp, argv[3], &refPoint, USE_PLACE_ONCE)
               == TCL_OK) {
      /* Reference point is a named place */
      rlat = refPoint.ord;
      rlon = refPoint.abs;
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0],
                       " projection LambertConfConic ?placeOR{refLat refLon} ",
                       NULL);
      return TCL_ERROR;
    }

    if (rlat == 0.0) {

      /*
       * Lambert Conformal Conic with reference latitude == equator
       * => Mercator projection, by defintion
       */
 
      SetMercator(proj, rlon);

    } else if (rlat == 90.0) {

      /* Lambert conformal conic -> polar stereographic at pole */
      SetPolarStereographic(proj, North);

    } else if (rlat == -90.0) {

      /* Lambert conformal conic -> polar stereographic at pole */
      SetPolarStereographic(proj, South);

    } else {

      /* Conventional Lambert conformal conic */
      SetLambertConfConic(proj, rlat, rlon);

    }

  } else if (   strcmp("PolarStereographic", argv[2]) == 0
             || strcmp("polarstereographic", argv[2]) == 0
             || strcmp("PolSterGrph", argv[2]) == 0
             || strcmp("PSG", argv[2]) == 0) {

    /* Set polar stereographic projection */
    enum LatSgn ns;

    if (argc != 4) {
      Tcl_AppendResult(interp, "Usage: ", argv[0], " projection "
        "polarstereographic N|S", NULL);
      return TCL_ERROR;
    }

    if (strcmp("N", argv[3]) == 0) {

      /* Arctic polar stereographic with 0 longitude vertical on map */
      ns = North;

    } else if (strcmp("S", argv[3]) == 0) {

      /* Antarctic polar stereographic with 0 longitude vertical on map */
      ns = South;

    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], " projection "
                       "PolarStereoGraphic ?90 or -90 refLon?", NULL);
      return TCL_ERROR;
    }
    geoMapPtr->center.abs = 0.0;
    geoMapPtr->center.ord = (ns == North) ? 90.0 : -90.0;
    SetPolarStereographic(proj, ns);

  } else if (   strcmp("sinusoidal", argv[2]) == 0
             || strcmp("Sinusoidal", argv[2]) == 0) {

    /* Set sinusoidal projection */
      
    if (argc == 3) {
      /* Use longitude of widget center as reference longitude */
      SetSinusoidalProj(geoMapPtr->projection, geoMapPtr->center.abs);
    } else if (argc == 4 && sscanf(argv[3], "%lf", &refLon) == 1) {
      /* Use given reference longitude */
      SetSinusoidalProj(geoMapPtr->projection, refLon);
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], " projection sinusoidal "
                       "?reflon?", NULL);
      return TCL_ERROR;
    }

  } else if (   strcmp("orthographic", argv[2]) == 0
             || strcmp("Orthographic", argv[2]) == 0) {
    
    /* Set orthographic projection */

    if (argc == 3) {
      /* Use widget center as reference point */
      SetOrthographic(geoMapPtr->projection, geoMapPtr->center);
    } else if (   argc == 4
               && (sscanf(argv[3], "%f %f",
                                      &refPoint.ord, &refPoint.abs) == 2
                   || GeoPlaceGet(interp, argv[3], &refPoint, USE_PLACE_ONCE)
	                          == TCL_OK)) {
      /* Use given reference point */
      SetOrthographic(geoMapPtr->projection, refPoint);
    } else {
      Tcl_AppendResult(interp, "Usage: ", argv[0], "projection orthographic "
                       "placeOR{refLat refLon}", NULL);
      return TCL_ERROR;
    }

  } else {
    Tcl_AppendResult(interp, "Unknown projection ", argv[2] ,
                     ".  Choices are: CylEqdist Mercator CylEqArea "
                     "LambertConfConic PolarStereographic Sinusoidal"
		     " Orthographic", NULL);
    return TCL_ERROR;
  }

  Tcl_SetResult(interp, ProjDescriptor(proj), TCL_VOLATILE);
  geoMapDraw(geoMapPtr);
  return TCL_OK;
}

int
rotation(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * rotation - set result to or change map rotation (degrees clockwise)
   * Usage:
   *   pathName rotation [degrees]
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Projection proj = geoMapPtr->projection;
  double rotngl;

  if (argc == 2) {
    sprintf(result, "%.3f degrees", GeoProjGetRotation(proj));
    Tcl_SetResult(interp, result, TCL_VOLATILE);
    return TCL_OK;
  }

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0],
                     " rotation ?degrees clockwise?", NULL);
    return TCL_ERROR;
  }

  if (sscanf(argv[2], "%lf", &rotngl) == 1) {
    GeoProjSetRotation(proj, rotngl);
  } else {
    Tcl_AppendResult(interp, argv[0], "Could not read rotation angle ",
      argv[2], NULL);
    return TCL_ERROR;
  }
  geoMapDraw(geoMapPtr);
  return TCL_OK;
}

int
latLonToProj(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* latLonToProj - convert latitude and longitude to projection
   * Usage:
   *   pathName latlontoproj {lat lon}
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;
  GeoPoint point;               /* Geographic coordinates of point */

  if (argc != 3 || sscanf(argv[2], "%f %f", &point.ord, &point.abs) != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0],
                     " latlontoproj {lat lon}", NULL);
    return TCL_ERROR;
  }

  point = LatLonToProj(point, geoMapPtr->projection);
  sprintf(result, "%.3f %.3f", point.ord, point.abs);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
projToLatLon(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* projToLatLon - convert projection to latitude and longitude
   * Usage:
   *   pathName projtolatlon {ordinate abscissa}
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  GeoPoint point;               /* Geographic coordinates of point */

  if (argc != 3 || sscanf(argv[2], "%f %f", &point.ord, &point.abs) != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0],
                     " projtolatlon {ordinate abscissa}", NULL);
    return TCL_ERROR;
  }

  point = ProjToLatLon(point, geoMapPtr->projection);
  sprintf(result, "%.3f %.3f", point.ord, point.abs);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
latLonToX(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Convert latitude and longitude to XPoint
   * Usage:
   *   pathName latlontox {lat lon}
   * Input:
   *   clientData - pointer to a tkgeomap
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  XPoint xPoint;		/* X-coordinates of point of interest */
  GeoPoint
    latLon,              	/* Geographic coordinates of point */
    projPt,			/* Projection coordinates of point */
    projCtr;                    /* Projection coordinates of center */

  if (argc != 3 || sscanf(argv[2], "%f %f", &latLon.ord, &latLon.abs) != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " latlontox {lat lon}", NULL);
    return TCL_ERROR;
  }

  /* Get widget projection coordinates of point */
  projPt = LatLonToProj(latLon, geoMapPtr->projection);

  /* Get widget X-coordinates of center */
  projCtr = LatLonToProj(geoMapPtr->center, geoMapPtr->projection);

  /* Convert to window X-coordinates */
  xPoint.x = geoMapPtr->width / 2 - (projCtr.abs - projPt.abs)*PxPerCm(interp);
  xPoint.y = geoMapPtr->height / 2 - (projPt.ord - projCtr.ord)*PxPerCm(interp);

  sprintf(result, "%hd %hd", xPoint.x, xPoint.y);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
xToLatLon(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* xToLatLon - convert XPoint in window to latitude and longitude
   * Usage:
   *   pathName xtolatlon {x y}
   * Input:
   *   clientData - pointer to a tkgeomap
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  short x, y;		/* X-coordinates of point of interest */
  GeoPoint
    latLon,		/* lat and lon of point of interest */
    projPt,		/* Projection coordinates of point */
    projCtr;		/* Projection coordinates of center */

  /* Read X-coordinates from the command line */
  if (argc != 3 || sscanf(argv[2], "%hd %hd", &x, &y) != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " xtolatlon {x y}", NULL);
    return TCL_ERROR;
  }

  projCtr = LatLonToProj(geoMapPtr->center, geoMapPtr->projection);
  projPt.abs = projCtr.abs - (geoMapPtr->width / 2 - x) * CmPerPx(interp);
  projPt.ord = projCtr.ord + (geoMapPtr->height / 2 - y) * CmPerPx(interp);
  latLon = ProjToLatLon(projPt, geoMapPtr->projection);
  sprintf(result, "%.3f %.3f", latLon.ord, latLon.abs);
  Tcl_SetResult(interp, result, TCL_VOLATILE);
  return TCL_OK;
}

int
shift(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* shift - move the widget pixmap by given delta-x and delta-y.
   * Fill newly exposed areas with widget background color.
   * Usage:
   *   pathname shift deltaX deltaY
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tk_Window tkwin = geoMapPtr->tkwin;
  int deltaX, deltaY;

  if (argc != 4) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " shift deltaX deltaY", NULL);
    return TCL_ERROR;
  } else {
    if (Tcl_GetInt(interp, argv[2], &deltaX) == TCL_ERROR ||
        Tcl_GetInt(interp, argv[3], &deltaY) == TCL_ERROR)
      return TCL_ERROR;

    shiftArea(Tk_Display(tkwin), Tk_WindowId(tkwin), geoMapPtr->gc,
              geoMapPtr->borderWidth, geoMapPtr->borderWidth,
              Tk_Width(tkwin), Tk_Height(tkwin), deltaX, deltaY,
              Tk_Depth(tkwin));
  }
  return TCL_OK;
}

int
drawLines(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Draw a set of lines in the projection
   * Usage:
   *   pathname drawlines lineArrayName ?-c color -h thickness -y style
   *     -d dotsize -b bitmap?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tk_Window tkwin = geoMapPtr->tkwin;
  char *lineArrayName;          		/* Name of the line array */
  GeoLineArray projLineArray = NULL;		/* Lines in projection */
  GeoMapLineArray *geoMapLineArray = NULL;	/* Data for widget lineArray */
  Tcl_HashEntry *entry;
  char **arg;                   		/* Loop index */
  int newPtr = 0;               		/* Tell if old array exists */
        
  /* Check the call */
  if (argc < 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " drawlines lineArrayName"
                     " ?-c color -h thickness -y style -d dotsize?", NULL);
    return TCL_ERROR;
  }

  lineArrayName = argv[2];

  /* Get data for the lineArray, adding it to the widget if necessary */
  entry = Tcl_CreateHashEntry(&geoMapPtr->lines, lineArrayName, &newPtr);

  if (newPtr) {

    /* Register as a user of the lineArray */
    if (GeoLatLonUseArray(lineArrayName) == TCL_ERROR) {
      Tcl_DeleteHashEntry(entry);
      Tcl_AppendResult(interp, argv[0], ": no line array named ",
        lineArrayName, NULL);
      return TCL_ERROR;
    }

    /* Create GeoLines in the widget */
    if ( !(geoMapLineArray = malloc(sizeof(GeoMapLineArray))) )
      goto error;

    /* Get projection coordinates of latLonArray */
    if ( !(projLineArray = LatLonArrayToProj(GetLatLonArray(lineArrayName),
                                             geoMapPtr->projection)) )
      goto error;

    /* Assign projLineArray to geoMapLineArray */
    geoMapLineArray->projLineArray = projLineArray;
        
    /* Set default attributes */
    geoMapLineArray->lineColor = NULL;
    geoMapLineArray->lineWidth = 0;
    geoMapLineArray->lineStyle = LineSolid;
    geoMapLineArray->dotSize = 0;
    geoMapLineArray->pixmap = None;
      
    Tcl_SetHashValue(entry, geoMapLineArray);
  } else {
    /* Get pointer to old array in widget */
    geoMapLineArray = Tcl_GetHashValue(entry);
  }

  if (argc > 3) {
    /* Parse options */
    for (arg = argv + 3; *arg; arg++) {
      if (strncmp(*arg, "-c", 2) == 0) {
        /* Set line color */
        XColor *color;
        Tk_Uid colorName;

        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1],
	    ": -c must have option", NULL);
          break;
        }
        if (strcmp(*arg, "default") == 0) {
          /* Use default color */
          geoMapLineArray->lineColor = NULL;
        } else {
          /* User selected line color */
          colorName = Tk_GetUid(*arg);
        
          if ((color = Tk_GetColor(interp, tkwin, colorName))) {
              if (geoMapLineArray->lineColor)
	        Tk_FreeColor(geoMapLineArray->lineColor);

              geoMapLineArray->lineColor = color;
          } else {
	    Tcl_AppendResult(interp, " ", argv[0], " ", argv[1],
	      " could not set color for ", lineArrayName, NULL);
	    break;
	  }
        }
      } else if (strncmp(*arg, "-h", 2) == 0) {
        /* Set line thickness */
        if (! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1],
	    ": -h must have option", NULL);
          break;
        }
        if (strcmp(*arg, "nolines") == 0) {
          /* Draw dots and bitmaps only */
          geoMapLineArray->lineWidth = -1;
        } else if (sscanf(*arg, "%d", &geoMapLineArray->lineWidth) != 1) {
          /* User selected line width */
          Tcl_AppendResult(interp, argv[0], " ", argv[1], " expected integer "
	    "linewidth for ", lineArrayName, ", got ", *arg, NULL);
        }
      } else if (strncmp(*arg, "-y", 2) == 0) {
        /* Set line style */
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1],
	    ": -y must have option", NULL);
          break;
        }
        if (strcmp(*arg, "LineSolid") == 0)
          geoMapLineArray->lineStyle = LineSolid;
        else if (strcmp(*arg, "LineOnOffDash") == 0)
          geoMapLineArray->lineStyle = LineOnOffDash;
        else if (strcmp(*arg, "LineDoubleDash") == 0)
          geoMapLineArray->lineStyle = LineDoubleDash;
        else
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": No style named",
	    *arg, "Choices are: LineSolid, LineOnOffDash, LineDoubleDash.",
	    NULL);
      } else if (strncmp(*arg, "-d", 2) == 0) {
        /* Draw a dot at end of each segment */
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1],
	    ": -d must have option", NULL);
          break;
        }
        if (strcmp(*arg, "default") == 0) {
          geoMapLineArray->dotSize = -1;
          continue;
        }
        if (sscanf(*arg, "%d", &geoMapLineArray->dotSize) != 1) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], "Expected integer, "
	    "got ", *arg, NULL);
        }
      } else if (strncmp(*arg, "-b", 2) == 0) {
        /* Draw a bitmap at end of each segment */
        char *bitmapName;
        Pixmap pixmap;
              
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -b must have "
	    "option", NULL);
          break;
        }
        bitmapName = *arg;
        if (strcmp(bitmapName, "none") == 0) {
          /* Erase bitmap at endpoints*/
          geoMapLineArray->pixmap = None;
          geoMapXDisplayWhenIdle(geoMapPtr);
          continue;
        }
        pixmap = Tk_GetBitmap(interp, tkwin, Tk_GetUid(bitmapName));
        if (pixmap == None) {
          continue;
        } else {
          geoMapLineArray->pixmap = pixmap;
        }
      } else {
        Tcl_AppendResult(interp, argv[0], ": unknown option: ", *arg, NULL);
      }
    }
  }
  geoMapXDisplayWhenIdle(geoMapPtr);
  return TCL_OK;
  
  error:
  Tcl_DeleteHashEntry(entry);
  free(geoMapLineArray);
  FreeGeoLineArray(projLineArray);
  if (newPtr)
    GeoLatLonFreeArray(lineArrayName);
  return TCL_ERROR;
}

int
eraseLines(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /*
   * eraseLines - Erase a set of lines
   * Usage:
   *   pathname eraselines lineArrayName
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  char *lineArrayName;
  Tcl_HashEntry *entry;
  GeoMapLineArray *geoMapLineArray;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " eraselines lineArrayName",
                     NULL);
    return TCL_ERROR;
  }
  lineArrayName = argv[2];
  entry = Tcl_FindHashEntry(&geoMapPtr->lines, lineArrayName);
  if ( !entry ) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], ": no line set named ",
      lineArrayName, NULL);
    return TCL_ERROR;
  }
  geoMapLineArray = Tcl_GetHashValue(entry);
  if (geoMapLineArray->lineColor) Tk_FreeColor(geoMapLineArray->lineColor);
  free(geoMapLineArray);
  Tcl_DeleteHashEntry(entry);
  GeoLatLonFreeArray(lineArrayName);
  geoMapXDisplayWhenIdle(geoMapPtr);
  return TCL_OK;
}

int
drawBitmap(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* drawBitmap - draw a bitmap at a named place
   * Usage:
   *   pathname drawbitmap placeName bitmapName ?color?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tk_Window tkwin = geoMapPtr->tkwin;
  char
    *placeName,
    *bitmapName;
  Pixmap pixmap;
  GeoMapPlace *place;           /* Data for place in widget's place table */

  /* Check the call */
  if (argc != 4 && argc != 5) {
    Tcl_AppendResult(interp, "Usage: ", argv[0],
                     " drawbitmap placeName bitmapName ?color?", NULL);
    return TCL_ERROR;
  }
  placeName = argv[2];
  bitmapName = argv[3];
  /* Get a pointer to the place */
  if ( !(place = geoMapCreatePlace(geoMapPtr, placeName)) ) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not find or "
      "allocate ", placeName, NULL);
    return TCL_ERROR;
  }

  if (strcmp(bitmapName, "none") == 0) {
    /* Remove the bitmap */
    place->pixmap = None;
    geoMapXDisplayWhenIdle(geoMapPtr);
    return TCL_OK;
  }

  pixmap = Tk_GetBitmap(interp, tkwin, Tk_GetUid(bitmapName));
  if (pixmap == None) {
    return TCL_ERROR;
  }
  place->pixmap = pixmap;

  if (argc == 4) {
    /* No optional color => use default color */
    place->bmColor = NULL;
  } else {
    /* Set optional color */
    Tk_Uid colorName = Tk_GetUid(argv[4]);
    XColor *color;

    if (strcmp(colorName, "default") == 0) {
      place->bmColor = NULL;
    } else if ((color = Tk_GetColor(interp, tkwin, colorName))) {
      if (place->bmColor) Tk_FreeColor(place->bmColor);
      place->bmColor = color;
    } else {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not apply "
        "bitmap color for ", placeName, NULL);
    }
  }
  geoMapXDisplayWhenIdle(geoMapPtr);

  return TCL_OK;
}

int
eraseBitmap(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Convenience function to erase a bitmap
   * Usage:
   *   pathName erasebitmap placeName
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " erasebitmap placeName",NULL);
    return TCL_ERROR;
  }
  if (geoMapCreatePlace(geoMapPtr, argv[2])) {
    return Tcl_VarEval(interp, argv[0], " drawbitmap ", argv[2], " none", NULL);
  }
  return TCL_OK;
}

int
drawDot(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Draw a dot at a place
   * Usage:
   *   pathName drawdot placeName ?-c color -s size?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  char **arg;                   /* Loop index */
  char *placeName;              /* Name of the place */
  GeoMapPlace *place;           /* Place widget data */

  /* Check the call */
  if (argc < 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " drawdot placeName"
                     " ?-c color -s size?", NULL);
    return TCL_ERROR;
  }

  placeName = argv[2];

  /* Get a pointer to the place */
  if ( !(place = geoMapCreatePlace(geoMapPtr, placeName)) ) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], "Could not find or "
      "allocate ", placeName, NULL);
    return TCL_ERROR;
  }

  if (argc == 3) {
    /* No options => set defaults */
    if (place->dotColor)
      Tk_FreeColor(place->dotColor);
    place->dotColor = NULL;
    place->dotSize = -1;
  } else {
    /* Parse options */
    for (arg = argv + 3; *arg; arg++) {
      if (strncmp(*arg, "-c", 2) == 0) {
        /* Set dot color */
        XColor *color;
        Tk_Window tkwin = geoMapPtr->tkwin;
        Tk_Uid colorName;

        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -c must have "
	    "option", NULL);
          break;
        }

        if (strcmp(*arg, "default") == 0) {
          /* Use default color */
          place->dotColor = NULL;
        } else {
          /* Use color specified on command line */
          colorName = Tk_GetUid(*arg);
        
          if ((color = Tk_GetColor(interp, tkwin, colorName))) {
            if (place->dotColor) Tk_FreeColor(place->dotColor);
            place->dotColor = color;
          }
        }
      } else if (strncmp(*arg, "-s", 2) == 0) {
        /* Set dot size */
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -s must have "
	    "option", NULL);
          break;
        }
        if (strcmp(*arg, "default") == 0) {
          place->dotSize = -1;
          continue;
        }
        if (sscanf(*arg, "%d", &place->dotSize) != 1) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], " expected integer, "
	    "got ", *arg, NULL);
	  break;
        }
      } else {
        Tcl_AppendResult(interp, argv[0], ": unknown option: ", *arg, NULL);
      }
    }  /* end for */
  }

  geoMapXDisplayWhenIdle(geoMapPtr);

  return TCL_OK;
}

int
eraseDot(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Convenience function to erase a dot (set its size to 0)
   * Usage:
   *   pathName erasedot placeName
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " erasedot placeName", NULL);
    return TCL_ERROR;
  }
  if (geoMapCreatePlace(geoMapPtr, argv[2])) {
    return Tcl_VarEval(interp, argv[0], " drawdot ", argv[2], " -s 0", NULL);
  }
  return TCL_OK;
}

int
text(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* text - write text at a named place
   * Usage:
   *   pathName text placeName "text..." ?-c color -fn font -a anchor?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  char
    *newText,                   /* Storage for new text */
    *placeName;                 /* Name of the place */
  GeoMapPlace *place;           /* Place widget data */
  char** arg;                   /* Loop index */

  /* Check the call */
  if (argc < 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0],
                     " text placeName \"text\" ?-c color -fn font "
                     "-a anchor?", NULL);
    return TCL_ERROR;
  }

  placeName = argv[2];

  /* Set result to the text being written at the place */
  if (argc == 3) {
    Tcl_HashEntry *entry;

    if ( !(entry = Tcl_FindHashEntry(&geoMapPtr->places, placeName)) ) {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], ": No place named ",
        placeName, NULL);
    }

    place = Tcl_GetHashValue(entry);
    Tcl_SetResult(interp, place->text, TCL_VOLATILE);

    return TCL_OK;
  }

  /* argc>3 => write or modify text */

  /* Get a pointer to the place */
  if ( !(place = geoMapCreatePlace(geoMapPtr, placeName)) ) {
    Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not find or "
      "allocate ", placeName, NULL);
    return TCL_ERROR;
  }

  /* Erase text */
  if (strlen(argv[3]) == 0) {
    place->text = NULL;
    geoMapXDisplayWhenIdle(geoMapPtr);
    return TCL_OK;
  } else if (strcmp(argv[3], "--") == 0) {
    /* No new text, just setting options */
  } else {
    /* Allocate storage for new text */
    if ( !(newText = (char *)malloc((strlen(argv[3]) + 1) * sizeof(char))) ) {
      Tcl_AppendResult(interp, argv[0], " ", argv[1], ": text allocation "
        "failed", NULL);
      return TCL_ERROR;
    }
        
    /* Copy the text */
    strcpy(newText, argv[3]);
        
    /* Free previous text, if any */
    if (place->text) free(place->text);
        
    /* Assign the new text */
    place->text = newText;
  }

  /* Parse options */
  if (argc > 4) {
    Tk_Window tkwin = geoMapPtr->tkwin; /* Widget window */

    for (arg = argv + 4; *arg; arg++) {
      if (strncmp(*arg, "-c", 2) == 0) {
        /* Set text color */
        XColor *color;
        Tk_Uid colorName;
        
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -c must have "
	    "option", NULL);
          break;
        }
        
        if (strcmp(*arg, "default") == 0) {
          /* Use default color */
          place->textColor = NULL;
        } else {
          /* Use place color */
          colorName = Tk_GetUid(*arg);
        
          if ( (color = Tk_GetColor(interp, tkwin, colorName)) ) {
            if (place->textColor) Tk_FreeColor(place->textColor);
            place->textColor = color;
          }
        }
      } else if (strcmp(*arg, "-fn") == 0 || strcmp(*arg, "-font") == 0) {
        /* Set text font */

        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -fn must have "
	    "option", NULL);
          break;
        }
        
        if (strcmp(*arg, "default") == 0) {
          /* Use widget default */
          place->font = NULL;
        } else {
          XFontStruct *font;
          char *fontName = *arg;

          if ((font = XLoadQueryFont(geoMapPtr->display, fontName))) {
            place->font = font;
          } else {
            Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not load "
	      "font ", fontName, NULL);
            break;
          }
        }
      } else if (strcmp(*arg, "-a") == 0) {
        /* Set text anchor */
        if ( ! *++arg) {
          Tcl_AppendResult(interp, argv[0], " ", argv[1], ": -a must have "
	    "option", NULL);
          break;
        }

        if (strcmp(*arg, "default") == 0) {
          /* Use widget default */
          place->textAnchor = NULL;
        } else {
          if ( !(place->textAnchor = (Tk_Anchor *)malloc(sizeof(Tk_Anchor))) ) {
            Tcl_AppendResult(interp, argv[0], " ", argv[1], " could not "
	      "allocate anchor", NULL);
            continue;
          }

          Tk_GetAnchor(interp, *arg, place->textAnchor);
        }
      } else {
        Tcl_AppendResult(interp, argv[0], "Unknown option: ", *arg, NULL);
      }
    }
  }

  geoMapXDisplayWhenIdle(geoMapPtr);
  return TCL_OK;
}

int
eraseText(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* Convenience function to erase text
   * Usage:
   *   pathName erasetext placeName
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " erasetext placeName", NULL);
    return TCL_ERROR;
  }

  if (geoMapCreatePlace(geoMapPtr, argv[2])) {
    return Tcl_VarEval(interp, argv[0], " Text ", argv[2], " \"\"", NULL);
  }
  return TCL_OK;
}

int
erase(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* erase - erase dot, text, and bitmap at named place and free up storage
   * Usage:
   *   pathName erase placeName
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  GeoMap *geoMapPtr = (GeoMap *)clientData;

  if (argc != 3) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " erase placeName.", NULL);
    return TCL_ERROR;
  }
  geoMapRmPlace(geoMapPtr, argv[2]);
  geoMapXDisplayWhenIdle(geoMapPtr);
  return TCL_OK;
}

int
listWdgPlaces(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
  /* listWdgPlaces - return a list of named places known to a widget
   * Usage:
   *   pathName listplaces
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  if (argc != 2) {
    Tcl_AppendResult(interp, "Usage: ", argv[0], " listplaces", NULL);
    return TCL_ERROR;
  }
  for (entry = Tcl_FirstHashEntry(&geoMapPtr->places, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    Tcl_AppendElement(interp, Tcl_GetHashKey(&geoMapPtr->places, entry));
  }
  return TCL_OK;
}

int
geoMapConfigure(Tcl_Interp *interp, GeoMap *geoMapPtr, int argc, char **argv,
		int flags)
{
  /* geoMapConfigure -set or return configuration options for a geomap
   * Usage:
   *   pathName configure ?option? ?value? ?option value ...?
   * Input:
   *   clientData - a Tk_Window token
   *   interp     - tcl interpreter
   *   argc, argv - the command line
   */

  GeoMap oldGeoMap = *geoMapPtr; /* Save geoMapPtr, in case of errors */

  if (Tk_ConfigureWidget(interp, geoMapPtr->tkwin, configSpecs,
                         argc, argv, (char *)geoMapPtr, flags) != TCL_OK)
    return TCL_ERROR;

  /*Set the background for the window and create a graphics context */
  Tk_SetWindowBackground(geoMapPtr->tkwin,
                         Tk_3DBorderColor(geoMapPtr->bgBorder)->pixel);

  /* Set graphics context for new widget */
  if (geoMapPtr->gc == None) {
    XGCValues gcValues;
    gcValues.function = GXcopy;
    gcValues.graphics_exposures = False;
    Tk_MakeWindowExist(geoMapPtr->tkwin);
    geoMapPtr->gc = XCreateGC(geoMapPtr->display, Tk_WindowId(geoMapPtr->tkwin),
                              GCFunction|GCGraphicsExposures, &gcValues);
  }

  /* Change font */
  if (geoMapPtr->fontName != oldGeoMap.fontName) {
    XFontStruct *font;
    if ((font = XLoadQueryFont(geoMapPtr->display, geoMapPtr->fontName))) {
      geoMapPtr->font = font;
    } else {
      Tcl_AppendResult(interp, argv[0], " could not load font: ",
        geoMapPtr->fontName, NULL);
    }
  }

  /* Register the desired geometry for the window. */
  Tk_GeometryRequest(geoMapPtr->tkwin, geoMapPtr->width, geoMapPtr->height);

  /* Change scale if necessary */
  if (geoMapPtr->scale != oldGeoMap.scale) {
    double
      numer, denom,		/* Use to calculate new scale */
      scale;			/* Scale read directly */
	  
    if (sscanf(geoMapPtr->scale, "%lf%*[/:]%lf", &numer, &denom) == 2) {

      /* Scale given as a ratio */
      GeoProjSetScale(geoMapPtr->projection, numer / denom);
      geoMapDraw(geoMapPtr);
      return TCL_OK;

    } else if (sscanf(geoMapPtr->scale, "%lf", &scale) == 1) {

      /* Scale given as one floating point number */
      char cScale[12];	/* Hold scale as string */
	  
      /* Set new scale */
      GeoProjSetScale(geoMapPtr->projection, scale);

      /* Write scale using cartographic convention */
      sprintf(cScale, "1:%-9ld", (long)(1.0 / scale));
      geoMapPtr->scale = Tk_GetUid(cScale);
	  
      /* Apply new scale to lines */
      geoMapDraw(geoMapPtr);
      return TCL_OK;

    } else {

      /* Scan failed, use previous scale */
      Tcl_AppendResult(geoMapPtr->interp, argv[0], ": ", geoMapPtr->scale,
        " not a scale\nShould be of form ddd:ddd or ddd/ddd", NULL);
      geoMapPtr->scale = oldGeoMap.scale;
    }

  }

  /* Make sure default dot size is non-negative */
  if (geoMapPtr->dotSize < 0) {
    Tcl_AppendResult(geoMapPtr->interp, "Dot size must not be negative", NULL);
    geoMapPtr->dotSize = oldGeoMap.dotSize;
  }

  geoMapXDisplayWhenIdle(geoMapPtr);
  return TCL_OK;
}

void
geoMapEventProc(ClientData clientData, XEvent *eventPtr)
{
  /* Handle certain events dispatched to a geomap widget
   * Input:
   *   clientData - pointer to a geomap widget
   *   eventPtr   - pointer to an event record */
  
  GeoMap *geoMapPtr = (GeoMap *) clientData;

  if (eventPtr->type == Expose) {
    geoMapXDisplayWhenIdle(geoMapPtr);
  } else if (eventPtr->type == ConfigureNotify) {   
    geoMapXDisplayWhenIdle(geoMapPtr);
  } else if (eventPtr->type == DestroyNotify) {
    Tcl_DeleteCommand(geoMapPtr->interp, Tk_PathName(geoMapPtr->tkwin));
    geoMapPtr->tkwin = NULL;
    if (geoMapPtr->updatePending)
      Tk_CancelIdleCall(geoMapXDisplay, (ClientData) geoMapPtr);
    Tk_EventuallyFree((ClientData) geoMapPtr, geoMapDestroy);
  }
}

void
geoMapDraw(ClientData clientData)
{
  /* geoMapDraw - apply projection to lines and places in geomap
   * widget pointed to by geoMapPtr and arrange to redraw it
   */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tcl_Interp *interp = geoMapPtr->interp;
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;
  GeoMapPlace *place;

  geoMapPtr->drawPending = 0;

  /* Apply the projection to the lines in the widget */
  for (entry = Tcl_FirstHashEntry(&geoMapPtr->lines, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    char *lineArrayName = Tcl_GetHashKey(&geoMapPtr->lines, entry);
    GeoMapLineArray *geoMapLineArray = Tcl_GetHashValue(entry);
    GeoLineArray projLineArray;

    if ( !(projLineArray = LatLonArrayToProj(GetLatLonArray(lineArrayName),
                                             geoMapPtr->projection)) ) {
      fprintf(stderr, "Could not apply projection to %s\n", lineArrayName);
      return;
    }
    FreeGeoLineArray(geoMapLineArray->projLineArray); /* Free previous array */
    geoMapLineArray->projLineArray = projLineArray;
  } /* end for */

  /* Apply the projection to the places in the widget */
  for (entry = Tcl_FirstHashEntry(&geoMapPtr->places, &search);
       entry != NULL; entry = Tcl_NextHashEntry(&search)) {
    char *placeName = Tcl_GetHashKey(&geoMapPtr->places, entry);
    GeoPoint latLon;
      
    if (GeoPlaceGet(interp, placeName, &latLon, USE_PLACE_ONCE) == TCL_ERROR) {
      fprintf(stderr, "Bug!  %s missing from database", placeName);
      return;
    }
    place = Tcl_GetHashValue(entry);
    place->projPoint = LatLonToProj(latLon, geoMapPtr->projection);
  }

  if (geoMapPtr->tkwin && Tk_IsMapped(geoMapPtr->tkwin))
    geoMapXDisplay(geoMapPtr);

  return;
}

void
geoMapXDisplay(ClientData clientData)
{
  /* Cycle through widget elements (lines and places) and invoke
   * graphics calls to draw dots, lines, bitmaps, and text
   * Input: clientData - pointer to geomap widget to draw
   */
  
  GeoMap *geoMapPtr = (GeoMap *) clientData;
  Tcl_Interp *interp = geoMapPtr->interp;
  Tk_Window tkwin = geoMapPtr->tkwin;
  Display *display = Tk_Display(tkwin);

  XGCValues
    gcVals,				/* Preserve graphics context */
    gcTmpVals;				/* Hold gc values to change */

  GeoMapLineArray *geoMapLineArray;	/* Line drawing data */
  GeoLineArray projLineArray;		/* Line array in projection coord's */
  GeoPoint projCenter;			/* Projection coordinates of center */
  XPoint *xPoints;			/* Points in one line */

  int i;				/* Loop indeces */
  double top, right, bottom, left;	/* Map edges */

  Pixmap wdgPM = None;			/* Drawing buffer */

  XColor
    *bgColor,				/* Background color */
    *color;				/* Foreground color */

  Tcl_HashEntry *entry;			/* For table searches */
  Tcl_HashSearch search;

  short x, y;				/* Window coordinates */
  short dotShift;			/* Shift to center dot */
  int dotCnt;				/* Pixels in a dot */
  XPoint *baseDot, *dot;		/* Hold points for dot drawing */
  int dotSize;				/* Size of dot to draw, in pixels */
  int spDotCnt;				/* Number of pixels in a special dot */
  int spShift;				/* Pixels to shift to center dot */
  XPoint *dotPtr, *endDot, *basePtr;	/* Loop pointers */
  int nl, np;				/* Loop indeces */

  /********************** End of declarations **********************/

  geoMapPtr->updatePending = 0;

  if (!Tk_IsMapped(tkwin))
    return;

  /* Preserve graphics context values */
  if ( !XGetGCValues(display, geoMapPtr->gc,
                     GCForeground|GCLineWidth|GCLineStyle, &gcVals) ) {
    fprintf(stderr, "%s: Could not preserve graphics context.\n",
            Tk_PathName(tkwin));
    return;
  }

  wdgPM = XCreatePixmap(display, Tk_WindowId(tkwin),
                        Tk_Width(tkwin), Tk_Height(tkwin),
                        DefaultDepthOfScreen(Tk_Screen(tkwin)));

  /* Fill the pixmap with background */
  bgColor = Tk_3DBorderColor(geoMapPtr->bgBorder);
  XSetForeground(display, geoMapPtr->gc, bgColor->pixel);
  XFillRectangle(display, wdgPM, geoMapPtr->gc,0, 0,
                 Tk_Width(tkwin), Tk_Height(tkwin));

  /* Set the foreground color for drawing map outlines */
  XSetForeground(display, geoMapPtr->gc, geoMapPtr->lineColor->pixel);

  /* Create "template" for faster drawing of default dot */
  dotCnt = geoMapPtr->dotSize * geoMapPtr->dotSize;
  dotShift = geoMapPtr->dotSize / 2;

  baseDot = (XPoint *)malloc(dotCnt * sizeof(XPoint));
  dot = (XPoint *)malloc(dotCnt * sizeof(XPoint));

  if ( !baseDot || !dot ) {
    fprintf(stderr, "%s: Could not allocate storage for dot\n",
            Tk_PathName(tkwin));
    return;
  }

  for (i = 0; i < dotCnt; i++) {
    baseDot[i].x = i % geoMapPtr->dotSize - dotShift;
    baseDot[i].y = i / geoMapPtr->dotSize - dotShift;
  }

  /* Draw the lines */

  /* Window limits in projection coordinates */
  projCenter = LatLonToProj(geoMapPtr->center, geoMapPtr->projection);
  top = projCenter.ord + geoMapPtr->height / (PxPerCm(interp) * 2.0);
  right = projCenter.abs + geoMapPtr->width / (PxPerCm(interp) * 2.0);
  bottom = projCenter.ord - geoMapPtr->height / (PxPerCm(interp) * 2.0);
  left = projCenter.abs - geoMapPtr->width / (PxPerCm(interp) * 2.0);

  for (entry = Tcl_FirstHashEntry(&geoMapPtr->lines, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    
    geoMapLineArray = Tcl_GetHashValue(entry);
    projLineArray = geoMapLineArray->projLineArray;

    if ( !(xPoints=malloc(GeoLineArrayNmax(projLineArray)*sizeof(XPoint))) ) {
      fprintf(stderr, "Could not allocate xPoints array\n");
      return;
    }

    /* Set graphics context for line drawing */
    color = geoMapLineArray->lineColor ? geoMapLineArray->lineColor
                                       : geoMapPtr->lineColor;
    XSetForeground(display, geoMapPtr->gc, color->pixel);
        
    /* Draw the xLine array */
    for (nl = 0; nl < GeoLineArrayNLines(projLineArray); nl++) {

      /* Draw lines with points in window */
      GeoLine geoLine = GetGeoLine(projLineArray, nl);
      if (   GeoLineOrdMax(geoLine) > bottom
          && GeoLineOrdMin(geoLine) < top
	  && GeoLineAbsMax(geoLine) > left
	  && GeoLineAbsMin(geoLine) < right) {

	int npts = GeoLineNPts(geoLine);

	/* Loop through points on line drawing optional dots and bitmaps */
	for (np = 0; np < npts; np++) {
          GeoPoint geoPoint = GetGeoPoint(geoLine, np);
	  XPoint xPt;

	  xPoints[np].y = (top - geoPoint.ord) * PxPerCm(interp);
	  xPoints[np].x = (geoPoint.abs - left) * PxPerCm(interp);
	  xPt = xPoints[np];
                  
          if (geoMapLineArray->pixmap != None) {

            /* Draw a bitmap */
            Pixmap pixmap = geoMapLineArray->pixmap;
            int width, height;
		      
            XSetBackground(display, geoMapPtr->gc, bgColor->pixel);
              Tk_SizeOfBitmap(display, pixmap, &width, &height);
            XCopyPlane(display, pixmap, wdgPM, geoMapPtr->gc, 0, 0,
              width, height, xPt.x - (width >> 1),
              xPt.y - (height >> 1), 0x01);

          }

          if (geoMapLineArray->dotSize != 0) {

            /* Draw a dot at segment endpoints */
            dotSize = geoMapLineArray->dotSize > 0 ? geoMapLineArray->dotSize
                                                   : geoMapPtr->dotSize;
            if (dotSize == 1) {
              XDrawPoint(display, wdgPM, geoMapPtr->gc, xPt.x, xPt.y);
            } else if (dotSize == geoMapPtr->dotSize) {
              /* Draw default dot with previously defined "template" */
			  
              for (dotPtr = dot, endDot = dotPtr + dotCnt, basePtr = baseDot;
                   dotPtr < endDot;
                   dotPtr++, basePtr++) {
                dotPtr->x = xPt.x + basePtr->x;
                dotPtr->y = xPt.y + basePtr->y;
              }
              XDrawPoints(display, wdgPM, geoMapPtr->gc, dot,
                          dotCnt, CoordModeOrigin);
            } else {
              /* Draw special dot */
              spDotCnt = dotSize * dotSize;
              spShift = dotSize >> 1;

              for (i = 0; i < spDotCnt; i++)
                XDrawPoint(display, wdgPM, geoMapPtr->gc, 
                           xPt.x + i % dotSize - spShift,
                           xPt.y + i / dotSize - spShift);
                        }
            } /* Finished with dot */
          } /* Finished dots and bitmaps on line */


          /* If nonnegative lineWidth, connect the dots */
          if (geoMapLineArray->lineWidth >= 0) {
            gcTmpVals.line_width = geoMapLineArray->lineWidth;
            gcTmpVals.line_style = geoMapLineArray->lineStyle;
            XChangeGC(display, geoMapPtr->gc,
                      GCLineWidth|GCLineStyle, &gcTmpVals);
            XDrawLines(display, wdgPM, geoMapPtr->gc, xPoints,
                       npts, CoordModeOrigin);
          }

        } /* Finished one line in lineArray */

      } /* Finished lineArray */

      free(xPoints);
    }

  /* Loop through the places table, drawing dots, bitmaps, and text */
  for (entry = Tcl_FirstHashEntry(&geoMapPtr->places, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    GeoMapPlace *place = Tcl_GetHashValue(entry);
    GeoPoint nowhere = NOWHERE;
        
    if (   place->projPoint.abs != nowhere.abs
        && place->projPoint.ord != nowhere.ord
	&& place->projPoint.abs >=  left && place->projPoint.abs <=  right
	&& place->projPoint.ord <= top && place->projPoint.ord >= bottom) {

      y = (top - place->projPoint.ord) * PxPerCm(interp);
      x = (place->projPoint.abs - left) * PxPerCm(interp);

      if (place->pixmap != None) {

        /* Draw a bitmap */
        Pixmap pixmap = place->pixmap;
        int width, height;

        color = place->bmColor ? place->bmColor : geoMapPtr->bmColor;
        XSetForeground(display, geoMapPtr->gc, color->pixel);
        XSetBackground(display, geoMapPtr->gc, bgColor->pixel);
        Tk_SizeOfBitmap(display, pixmap, &width, &height);
                        XCopyPlane(display, pixmap, wdgPM, geoMapPtr->gc, 0, 0,
			width, height, x - width/2, y - height/2, 0x01);
      }

      if (place->dotSize != 0) {
        dotSize = place->dotSize > 0 ? place->dotSize
                                     : geoMapPtr->dotSize;

        color = place->dotColor ? place->dotColor : geoMapPtr->dotColor;
        XSetForeground(display, geoMapPtr->gc, color->pixel);

        if (dotSize == 1) {
          XDrawPoint(display, wdgPM, geoMapPtr->gc, x, y);
        } else if (dotSize == geoMapPtr->dotSize) {
          /* Draw default dot */
          for (dotPtr = dot, endDot = dotPtr + dotCnt, basePtr = baseDot;
               dotPtr < endDot;
               dotPtr++, basePtr++) {
            dotPtr->x = x + basePtr->x;
            dotPtr->y = y + basePtr->y;
          }
          XDrawPoints(display, wdgPM, geoMapPtr->gc, dot, dotCnt,
                      CoordModeOrigin);
        } else {
          /* Draw special dot */
          spDotCnt = dotSize * dotSize;
          spShift = dotSize >> 1;

          for (i = 0; i < spDotCnt; i++) {
            XDrawPoint(display, wdgPM, geoMapPtr->gc, 
                       x + i % dotSize - spShift,
                       y + i / dotSize - spShift);
          }
        }
      }

      if (place->text) {
        XFontStruct *font;
        Tk_Anchor textAnchor;
        short xText = x, yText = y;
        
        color = place->textColor ? place->textColor
                                 : geoMapPtr->textColor;
        XSetForeground(display, geoMapPtr->gc, color->pixel);
        
        font = place->font ? place->font : geoMapPtr->font;
        XSetFont(display, geoMapPtr->gc, font->fid);
                
        /* Adjust for anchoring */
        textAnchor = place->textAnchor ? *place->textAnchor
                                       : geoMapPtr->textAnchor;

        if (   textAnchor == TK_ANCHOR_N
            || textAnchor == TK_ANCHOR_CENTER
            || textAnchor == TK_ANCHOR_S)
          xText -= XTextWidth(font, place->text, strlen(place->text)) >> 1;
        else if (   textAnchor == TK_ANCHOR_NW
                 || textAnchor == TK_ANCHOR_W
                 || textAnchor == TK_ANCHOR_SW)
          xText -= XTextWidth(font, place->text, strlen(place->text));
              
        if (   textAnchor == TK_ANCHOR_W
            || textAnchor == TK_ANCHOR_CENTER
            || textAnchor == TK_ANCHOR_E)
          yText += font->ascent >> 1;
        else if (   textAnchor == TK_ANCHOR_SW
                 || textAnchor == TK_ANCHOR_S
                 || textAnchor == TK_ANCHOR_SE)
          yText += font->ascent;
              
        XDrawString(display, wdgPM, geoMapPtr->gc, xText, yText,
                    place->text, strlen(place->text));
      }
    }
  }


  /* Draw the borders */
  Tk_Draw3DRectangle(tkwin, wdgPM, geoMapPtr->bgBorder, 0, 0,
                     Tk_Width(tkwin), Tk_Height(tkwin), geoMapPtr->borderWidth,
                     geoMapPtr->relief);
        
  /* Send the buffer pixmap to the screen */
  XCopyArea(display, wdgPM, Tk_WindowId(tkwin), geoMapPtr->gc, 0, 0,
            Tk_Width(tkwin), Tk_Height(tkwin), 0, 0);
  XFreePixmap(display, wdgPM);

  /* Restore foreground and background */
  XChangeGC(display, geoMapPtr->gc, 
            GCForeground|GCLineWidth|GCLineStyle, &gcVals);

  /* Free memory */
  free(baseDot);
  free(dot);
}

void
geoMapXDisplayWhenIdle(GeoMap *geoMapPtr)
{
  /* Request display of geoMapPtr next time tk not busy, if not already
   * requested */
  
  if (   geoMapPtr->tkwin && Tk_IsMapped(geoMapPtr->tkwin)
      && !geoMapPtr->updatePending) {
    Tk_DoWhenIdle(geoMapXDisplay, (ClientData)geoMapPtr);
    geoMapPtr->updatePending = 1;
  }
}

void
geoMapDestroy(char *clientData)
{
  /* Delete a geomap, associated data structures, and free memory
   * Input:  clientData - pointer to geomap widget */
  
  GeoMap *geoMapPtr = (GeoMap *)clientData;
  Tcl_Interp *interp = geoMapPtr->interp;
  Tcl_HashEntry *entry;
  Tcl_HashSearch search;

  Tk_FreeOptions(configSpecs, (char *)geoMapPtr, geoMapPtr->display, 0);

  if (Tcl_VarEval(interp, "geoplace updatecmd rm ",
                  geoMapPtr->name, NULL) == TCL_ERROR) {
    Tcl_AppendResult(interp, "Warning: Could not remove update command for ",
      geoMapPtr->name, NULL);
  }

  if (geoMapPtr->gc != None)
    XFreeGC(geoMapPtr->display, geoMapPtr->gc);

  for (entry = Tcl_FirstHashEntry(&geoMapPtr->places, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    geoMapRmPlace(geoMapPtr, Tcl_GetHashKey(&geoMapPtr->places, entry));
  }
  Tcl_DeleteHashTable(&geoMapPtr->places);

  for (entry = Tcl_FirstHashEntry(&geoMapPtr->lines, &search);
       entry != NULL;
       entry = Tcl_NextHashEntry(&search)) {
    GeoMapLineArray *mapLines;
    char *arrayName;

    mapLines = Tcl_GetHashValue(entry);
    if( mapLines ) {
      FreeGeoLineArray(mapLines->projLineArray);
      free(mapLines);
    }
    arrayName = Tcl_GetHashKey(&geoMapPtr->lines, entry);
    GeoLatLonFreeArray(arrayName);
  }
  Tcl_DeleteHashTable(&geoMapPtr->lines);

  FreeProj(geoMapPtr->projection);

  free(geoMapPtr);
}

GeoMapPlace *
geoMapCreatePlace(GeoMap *geoMapPtr, char *placeName)
{
  /* Return a pointer to a place on a GeoMap widget, creating it in the
   * widget if necessary.
   * Input:
   *   geoMapPtr - pointer to a geomap widget
   *   placeName - name of the place.  Should already have a location
   *     accessible with GeoPlaceGet.  If placeName is not already in
   *     the widget's table of places, a new place is added.
   * Return value:
   *   Pointer to value in the widget's table of places - a struct containing
   *     information about what to draw and write at that place.
   */

  Tcl_Interp *interp = geoMapPtr->interp;
  Tcl_HashEntry *entry;
  GeoMapPlace *place;           /* Data for place in widget */
  int newPtr;

  /* Get data for the place, creating it if necessary */
  entry = Tcl_CreateHashEntry(&geoMapPtr->places, placeName, &newPtr);

  if (newPtr == 1) {
    /* Allocate memory for new place */
    if ( !(place = (GeoMapPlace *)malloc(sizeof(GeoMapPlace))) ) {
      Tcl_AppendResult(geoMapPtr->interp, "Could not allocate storage for ",
                       placeName, NULL);
      Tcl_DeleteHashEntry(entry);
      return NULL;
    }
      
    /* Get location in master database */
    if (GeoPlaceGet(interp, placeName, &place->latLon, KEEP_PLACE)
	  == TCL_ERROR) {
      free(place);
      Tcl_DeleteHashEntry(entry);
      return NULL;
    }

    /* Initialize to defaults and add to table */
    place->dotColor = NULL;
    place->dotSize = 0;
    place->pixmap = None;
    place->bmColor = NULL;
    place->text = NULL;
    place->textColor = NULL;
    place->font = NULL;
    place->textAnchor = NULL;

    Tcl_SetHashValue(entry, place);
  } else {
    place = Tcl_GetHashValue(entry);
    if (GeoPlaceGet(interp, placeName, &place->latLon, USE_PLACE_ONCE)
	  == TCL_ERROR) {
      free(place);
      Tcl_DeleteHashEntry(entry);
      return NULL;
    }
  }

  /* Create or update X-coordinates of the place */
  place->projPoint = LatLonToProj(place->latLon, geoMapPtr->projection);

  return place;
}

void
geoMapRmPlace(GeoMap *geoMapPtr, char *placeName)
{
  /* geoMapRmPlace - free up storage and delete a place identified by
   * placeName from geomap widget pointed to by placeName */
  
  Tcl_HashEntry *entry;
  GeoMapPlace *place;           /* Data for place in widget */

  /* Decrement user count in master place table */
  GeoPlaceFree(placeName);

  /* Get entry for place in widget place table */
  if ( !(entry = Tcl_FindHashEntry(&geoMapPtr->places, placeName)) )
    return;

  place = Tcl_GetHashValue(entry);

  if( place ) {
    /* Free storage and remove from table */
    if (place->dotColor) Tk_FreeColor(place->dotColor);
    if (place->bmColor) Tk_FreeColor(place->bmColor);
    free(place->text);
    if (place->textColor) Tk_FreeColor(place->textColor);
    free(place);
  }

  Tcl_DeleteHashEntry(entry);

  return;
}

void
shiftArea(Display *display, Drawable mvPic, GC gc, int x, int y,
	  unsigned width, unsigned height, int deltaX, int deltaY,
	  unsigned depth)
{
  /* shiftArea - move a region of a pixmap in a window, leaving "unknown" 
   * areas blank (fill with background color).  Use if filling in the
   * the "unknown" areas would take too long.
   * Input:
   *   display - X display
   *   mvPic   - pixmap where drawing takes place
   *   x, y    - X coordinates of upper left corner of region to shift
   *   width   - width (in pixels) of region to shift
   *   height  - height (in pixels) of region to shift
   *   deltaX  - how far to the right (in pixels) to shift the region
   *   deltaY  - how far down (in pixels) to shift the region
   */
  
  XGCValues xgcValues;
  Pixmap tmpPm = XCreatePixmap(display, mvPic, width, height, depth);
  unsigned
    aDeltaX = deltaX > 0 ? deltaX : -deltaX,
    aDeltaY = deltaY > 0 ? deltaY : -deltaY;

  XGetGCValues(display, gc, GCGraphicsExposures, &xgcValues);
  XSetGraphicsExposures(display, gc, False);

  XCopyArea(display, mvPic, tmpPm, gc, x, y, width-2*x, height-2*y, 0, 0);
        
  XCopyArea(display, tmpPm, mvPic, gc,
            deltaX > 0 ? 0 : aDeltaX,
            deltaY > 0 ? 0 : aDeltaY, 
            width - 2 * x - aDeltaX, height - 2 * y - aDeltaY,
            deltaX > 0 ? x + deltaX : x,
            deltaY > 0 ? y + deltaY : y);

  if (deltaX != 0)
    XClearArea(display, mvPic,
               deltaX > 0 ? x : width - x - aDeltaX, y,
               aDeltaX, height - 2 * y, False);

  if (deltaY != 0)
    XClearArea(display, mvPic,
               x, deltaY > 0 ? y : height - y - aDeltaY,
               width - 2 * x, aDeltaY, False);

  XFreePixmap(display, tmpPm);
  XChangeGC(display, gc, GCGraphicsExposures, &xgcValues);
}
